"""
This file demonstrates a bokeh applet, which can either be viewed
directly on a bokeh-server, or embedded into a flask application.
See the README.md file in this directory for instructions on running.
"""

import logging

logging.basicConfig(level=logging.DEBUG)

from os import listdir
from os.path import dirname, join, splitext

import numpy as np
import pandas as pd

from bokeh.models import ColumnDataSource, Plot
from bokeh.plotting import figure, curdoc
from bokeh.properties import String, Instance
from bokeh.server.app import bokeh_app
from bokeh.server.utils.plugins import object_page
from bokeh.models.widgets import (HBox, VBox, VBoxForm, PreText,
                                  Select, AppHBox, AppVBox, AppVBoxForm)



from bokeh.simpleapp import simpleapp

select1 = Select(name='ticker1', value='AAPL', options=['AAPL', 'GOOG', 'INTC', 'BRCM', 'YHOO'])
select2 = Select(name='ticker2', value='GOOG', options=['AAPL', 'GOOG', 'INTC', 'BRCM', 'YHOO'])

@simpleapp(select1, select2)
def stock(ticker1, ticker2):
    pretext = PreText(text="", width=500)
    df = get_data(ticker1, ticker2)
    source = ColumnDataSource(data=df)
    source.tags = ['main_source']
    p = figure(
        title="%s vs %s" % (ticker1, ticker2),
        plot_width=400, plot_height=400,
        tools="pan,wheel_zoom,box_select,reset",
        title_text_font_size="10pt",
    )
    p.circle(ticker1 + "_returns", ticker2 + "_returns",
             size=2,
             nonselection_alpha=0.02,
             source=source
    )
    stats = df.describe()
    pretext.text = str(stats)
    row1 = HBox(children=[p, pretext])
    hist1 = hist_plot(df, ticker1)
    hist2 = hist_plot(df, ticker2)
    row2 = HBox(children=[hist1, hist2])
    line1 = line_plot(ticker1, source)
    line2 = line_plot(ticker2, source, line1.x_range)
    output =  VBox(children=[row1, row2, line1, line2])
    return output

stock.route("/bokeh/stocks/")

@simpleapp(select1, select2)
def stock2(ticker1, ticker2):
    pretext = PreText(text="", width=500)
    df = get_data(ticker1, ticker2)
    source = ColumnDataSource(data=df)
    source.tags = ['main_source']
    p = figure(
        title="%s vs %s" % (ticker1, ticker2),
        plot_width=400, plot_height=400,
        tools="pan,wheel_zoom,box_select,reset",
        title_text_font_size="10pt",
    )
    p.circle(ticker1 + "_returns", ticker2 + "_returns",
             size=2,
             nonselection_alpha=0.02,
             source=source
    )
    stats = df.describe()
    pretext.text = str(stats)
    hist1 = hist_plot(df, ticker1)
    hist2 = hist_plot(df, ticker2)
    line1 = line_plot(ticker1, source)
    line2 = line_plot(ticker2, source, line1.x_range)
    return dict(scatterplot=p,
                statstext=pretext,
                hist1=hist1,
                hist2=hist2,
                line1=line1,
                line2=line2)

@stock2.layout
def stock2_layout(app):
    widgets = AppVBoxForm(app=app, children=['ticker1', 'ticker2'])
    row1 = AppHBox(app=app, children=['scatterplot', 'statstext'])
    row2 = AppHBox(app=app, children=['hist1', 'hist2'])
    all_plots = AppVBox(app=app, children=[row1, row2, 'line1', 'line2'])
    app = AppHBox(app=app, children=[widgets, all_plots])
    return app

@stock2.update(['ticker1', 'ticker2'])
def stock2_update_input(ticker1, ticker2, app):
    return stock2(ticker1, ticker2)

@stock2.update([({'tags' : 'main_source'}, ['selected'])])
def stock2_update_selection(ticker1, ticker2, app):
    source = app.select_one({'tags' : 'main_source'})
    df = get_data(ticker1, ticker2)
    if source.selected:
        selected_df = df.iloc[source.selected['1d']['indices'], :]
    else:
        selected_df = df

    stats_text = app.objects['statstext']
    stats_text.text = str(selected_df.describe())
    return {
        'hist1': hist_plot(df, ticker1, selected_df=selected_df),
        'hist2': hist_plot(df, ticker2, selected_df=selected_df),
        'statstext': stats_text,
    }

stock2.route("/bokeh/stocks2/")

def hist_plot(df, ticker, selected_df=None):
    if selected_df is None:
        selected_df = df
    global_hist, global_bins = np.histogram(df[ticker + "_returns"], bins=50)
    hist, bins = np.histogram(selected_df[ticker + "_returns"], bins=50)
    width = 0.7 * (bins[1] - bins[0])
    center = (bins[:-1] + bins[1:]) / 2
    start = global_bins.min()
    end = global_bins.max()
    top = hist.max()
    p = figure(
        title="%s hist" % ticker,
        plot_width=500, plot_height=200,
        tools="",
        title_text_font_size="10pt",
        x_range=[start, end],
        y_range=[0, top],
    )
    p.rect(center, hist / 2.0, width, hist)
    return p

def line_plot(ticker, source, x_range=None):
    p = figure(
        title=ticker,
        x_range=x_range,
        x_axis_type='datetime',
        plot_width=1000, plot_height=200,
        title_text_font_size="10pt",
        tools="pan,wheel_zoom,box_select,reset"
    )
    p.circle(
        'date', ticker,
        size=2,
        source=source,
        nonselection_alpha=0.02
    )
    return p

# build up list of stock data in the daily folder
data_dir = join(dirname(__file__), "daily")
try:
    tickers = listdir(data_dir)
except OSError as e:
    print('Stock data not available, see README for download instructions.')
    raise e
tickers = [splitext(x)[0].split("table_")[-1] for x in tickers]

# cache stock data as dict of pandas DataFrames
pd_cache = {}


def get_ticker_data(ticker):
    fname = join(data_dir, "table_%s.csv" % ticker.lower())
    data = pd.read_csv(
        fname,
        names=['date', 'foo', 'o', 'h', 'l', 'c', 'v'],
        header=False,
        parse_dates=['date']
    )
    data = data.set_index('date')
    data = pd.DataFrame({ticker: data.c, ticker + "_returns": data.c.diff()})
    return data


def get_data(ticker1, ticker2):
    if pd_cache.get((ticker1, ticker2)) is not None:
        return pd_cache.get((ticker1, ticker2))

    # only append columns if it is the same ticker
    if ticker1 != ticker2:
        data1 = get_ticker_data(ticker1)
        data2 = get_ticker_data(ticker2)
        data = pd.concat([data1, data2], axis=1)
    else:
        data = get_ticker_data(ticker1)

    data = data.dropna()
    pd_cache[(ticker1, ticker2)] = data
    return data



# class StockApp(VBox):
#     extra_generated_classes = [["StockApp", "StockApp", "VBox"]]
#     jsmodel = "VBox"

#     # text statistics
#     pretext = Instance(PreText)

#     # plots
#     plot = Instance(Plot)
#     line_plot1 = Instance(Plot)
#     line_plot2 = Instance(Plot)
#     hist1 = Instance(Plot)
#     hist2 = Instance(Plot)

#     # data source
#     source = Instance(ColumnDataSource)

#     # layout boxes
#     mainrow = Instance(HBox)
#     histrow = Instance(HBox)
#     statsbox = Instance(VBox)

#     # inputs
#     ticker1 = String(default="AAPL")
#     ticker2 = String(default="GOOG")
#     ticker1_select = Instance(Select)
#     ticker2_select = Instance(Select)
#     input_box = Instance(VBoxForm)

#     def __init__(self, *args, **kwargs):
#         super(StockApp, self).__init__(*args, **kwargs)
#         self._dfs = {}

#     @classmethod
#     def create(cls):
#         """
#         This function is called once, and is responsible for
#         creating all objects (plots, datasources, etc)
#         """
#         # create layout widgets
#         obj = cls()

#         # create input widgets
#         obj.make_inputs()

#         # outputs
#         obj.pretext = PreText(text="", width=500)
#         obj.make_source()
#         obj.make_plots()
#         obj.make_stats()

#         # layout
#         obj.set_children()
#         return obj

#     def make_inputs(self):

#         self.ticker1_select = Select(
#             name='ticker1',
#             value='AAPL',
#             options=['AAPL', 'GOOG', 'INTC', 'BRCM', 'YHOO']
#         )
#         self.ticker2_select = Select(
#             name='ticker2',
#             value='GOOG',
#             options=['AAPL', 'GOOG', 'INTC', 'BRCM', 'YHOO']
#         )

#     @property
#     def selected_df(self):
#         pandas_df = self.df
#         selected = self.source.selected
#         if selected:
#             pandas_df = pandas_df.iloc[selected, :]
#         return pandas_df

#     def make_source(self):
#         self.source = ColumnDataSource(data=self.df)

#     def line_plot(self, ticker, x_range=None):
#         p = figure(
#             title=ticker,
#             x_range=x_range,
#             x_axis_type='datetime',
#             plot_width=1000, plot_height=200,
#             title_text_font_size="10pt",
#             tools="pan,wheel_zoom,box_select,reset"
#         )
#         p.circle(
#             'date', ticker,
#             size=2,
#             source=self.source,
#             nonselection_alpha=0.02
#         )
#         return p

#     def hist_plot(self, ticker):
#         global_hist, global_bins = np.histogram(self.df[ticker + "_returns"], bins=50)
#         hist, bins = np.histogram(self.selected_df[ticker + "_returns"], bins=50)
#         width = 0.7 * (bins[1] - bins[0])
#         center = (bins[:-1] + bins[1:]) / 2
#         start = global_bins.min()
#         end = global_bins.max()
#         top = hist.max()

#         p = figure(
#             title="%s hist" % ticker,
#             plot_width=500, plot_height=200,
#             tools="",
#             title_text_font_size="10pt",
#             x_range=[start, end],
#             y_range=[0, top],
#         )
#         p.rect(center, hist / 2.0, width, hist)
#         return p

#     def make_plots(self):
#         ticker1 = self.ticker1
#         ticker2 = self.ticker2
#         p = figure(
#             title="%s vs %s" % (ticker1, ticker2),
#             plot_width=400, plot_height=400,
#             tools="pan,wheel_zoom,box_select,reset",
#             title_text_font_size="10pt",
#         )
#         p.circle(ticker1 + "_returns", ticker2 + "_returns",
#                  size=2,
#                  nonselection_alpha=0.02,
#                  source=self.source
#         )
#         self.plot = p

#         self.line_plot1 = self.line_plot(ticker1)
#         self.line_plot2 = self.line_plot(ticker2, self.line_plot1.x_range)
#         self.hist_plots()

#     def hist_plots(self):
#         ticker1 = self.ticker1
#         ticker2 = self.ticker2
#         self.hist1 = self.hist_plot(ticker1)
#         self.hist2 = self.hist_plot(ticker2)

#     def set_children(self):
#         self.children = [self.mainrow, self.histrow, self.line_plot1, self.line_plot2]
#         self.mainrow.children = [self.input_box, self.plot, self.statsbox]
#         self.input_box.children = [self.ticker1_select, self.ticker2_select]
#         self.histrow.children = [self.hist1, self.hist2]
#         self.statsbox.children = [self.pretext]

#     def input_change(self, obj, attrname, old, new):
#         if obj == self.ticker2_select:
#             self.ticker2 = new
#         if obj == self.ticker1_select:
#             self.ticker1 = new

#         self.make_source()
#         self.make_plots()
#         self.set_children()
#         curdoc().add(self)

#     def setup_events(self):
#         super(StockApp, self).setup_events()
#         if self.source:
#             self.source.on_change('selected', self, 'selection_change')
#         if self.ticker1_select:
#             self.ticker1_select.on_change('value', self, 'input_change')
#         if self.ticker2_select:
#             self.ticker2_select.on_change('value', self, 'input_change')

#     def make_stats(self):
#         stats = self.selected_df.describe()
#         self.pretext.text = str(stats)

#     def selection_change(self, obj, attrname, old, new):
#         self.make_stats()
#         self.hist_plots()
#         self.set_children()
#         curdoc().add(self)

#     @property
#     def df(self):
#         return get_data(self.ticker1, self.ticker2)


# # The following code adds a "/bokeh/stocks/" url to the bokeh-server. This URL
# # will render this StockApp. If you don't want serve this applet from a Bokeh
# # server (for instance if you are embedding in a separate Flask application),
# # then just remove this block of code.
# @bokeh_app.route("/bokeh/stocks/")
# @object_page("stocks")
# def make_object():
#     app = StockApp.create()
#     return app



from __future__ import print_function

from numpy import pi, arange, sin
import numpy as np
import time

from bokeh.browserlib import view
from bokeh.document import Document
from bokeh.embed import file_html
from bokeh.models.glyphs import Circle
from bokeh.models import (
    Plot, DataRange1d, DatetimeAxis,
    ColumnDataSource, PanTool, WheelZoomTool
)
from bokeh.resources import INLINE

x = arange(-2 * pi, 2 * pi, 0.1)
y = sin(x)

# Create an array of times, starting at the current time, and extending
# for len(x) number of hours.
times = np.arange(len(x)) * 3600000 + time.time()

source = ColumnDataSource(
    data=dict(x=x, y=y, times=times)
)

xdr = DataRange1d()
ydr = DataRange1d()

plot = Plot(x_range=xdr, y_range=ydr, min_border=80)

circle = Circle(x="times", y="y", fill_color="red", size=5, line_color="black")
plot.add_glyph(source, circle)

plot.add_layout(DatetimeAxis(), 'below')
plot.add_layout(DatetimeAxis(), 'left')

plot.add_tools(PanTool(), WheelZoomTool())

doc = Document()
doc.add(plot)

if __name__ == "__main__":
    filename = "dateaxis.html"
    with open(filename, "w") as f:
        f.write(file_html(doc, INLINE, "Date Axis Example"))
    print("Wrote %s" % filename)
    view(filename)



from __future__ import absolute_import

import unittest

from bokeh.util.serialization import get_json, json_apply, make_id, urljoin, traverse_data

class DummyRequestCallable():
    def json(self):
        return True

class DummyRequestProperty():
    json = True

class TestMakeId(unittest.TestCase):
    def test_basic(self):
        self.assertEqual(len(make_id()), 36)
        self.assertTrue(isinstance(make_id(), str))

    def test_simple_ids(self):
        import os
        os.environ["BOKEH_SIMPLE_IDS"] = "yes"
        self.assertEqual(make_id(), "1001")
        self.assertEqual(make_id(), "1002")
        del os.environ["BOKEH_SIMPLE_IDS"]

class TestUrlJoin(unittest.TestCase):
    def test_urljoin(self):
        result1 = urljoin('http://www.bokeh.com', 'test/')
        self.assertEqual(result1, 'http://www.bokeh.com/test/')
        result2 = urljoin('http://www.bokeh.com', 'test1/', 'test2/',
                          'test3/', 'bokeh.html')
        self.assertEqual(result2, 'http://www.bokeh.com/test1/test2/test3/bokeh.html')
        result3 = urljoin('http://www.notbokeh.com', 'http://www.bokeh.com/',
                          'test1/', 'bokeh1.squig')
        self.assertEqual(result3, 'http://www.bokeh.com/test1/bokeh1.squig')

class TestGetJson(unittest.TestCase):
    def test_with_property(self):
        self.assertTrue(get_json(DummyRequestProperty()))

    def test_with_method(self):
        self.assertTrue(get_json(DummyRequestCallable()))

class TestJsonapply(unittest.TestCase):

    def test_jsonapply(self):

        def check_func(frag):
            if frag == 'goal':
                return True

        def func(frag):
            return frag + 'ed'

        result = json_apply('goal', check_func, func)
        self.assertEqual(result, 'goaled')
        result = json_apply([[['goal', 'junk'], 'junk', 'junk']], check_func, func)
        self.assertEqual(result, [[['goaled', 'junk'], 'junk', 'junk']])
        result = json_apply({'1': 'goal', 1.5: {'2': 'goal', '3': 'junk'}}, check_func, func)
        self.assertEqual(result, {'1': 'goaled', 1.5: {'2': 'goaled', '3': 'junk'}})


class TestTraverseData(unittest.TestCase):
    testing = [[float('nan'), 3], [float('-inf'), [float('inf')]]]
    expected = [['NaN', 3.0], ['-Infinity', ['Infinity']]]

    def test_return_valid_json(self):
        self.assertTrue(traverse_data(self.testing) == self.expected)

    def test_with_numpy(self):
        self.assertTrue(traverse_data(self.testing, True) == self.expected)

    def test_without_numpy(self):
        self.assertTrue(traverse_data(self.testing, False) == self.expected)

if __name__ == "__main__":
    unittest.main()



""" Properties for modeling Chart inputs, constraints, and dependencies.

selection spec:
    [['x'], ['x', 'y']]
    [{'x': categorical, 'y': numerical}]

"""

from __future__ import absolute_import
import numpy as np
import pandas as pd

from bokeh.properties import (HasProps, Either, String, Int, List, Bool,
                              PrimitiveProperty, bokeh_integer_types, Array)

from .utils import special_columns, title_from_columns


class Column(Array):

    def _is_seq(self, value):
        is_array = super(Column, self)._is_seq(value)
        return isinstance(value, pd.Series) or isinstance(value, list) or is_array

    def _new_instance(self, value):
        return pd.Series(value)

    def transform(self, value):
        if value is None:
            return None

        if isinstance(value, pd.Series):
            arr = value.values
        else:
            arr = value

        trans_array = super(Column, self).transform(arr)
        try:
            return pd.Series(trans_array)
        except ValueError:

            raise ValueError("Could not transform %r" % value)


class Logical(Bool):
    """A boolean like data type."""
    def validate(self, value):
        try:
            super(Logical, self).validate(value)
        except ValueError:
            if isinstance(value, list):
                value = np.array(value)

            # If not a Bool, then look for psuedo-logical types
            if isinstance(value, np.ndarray):
                values = np.unique(value)
                if len(values) == 2:
                    return

                raise ValueError('expected a Bool or array with 2 unique values, got %s' % value)


class ColumnLabel(Either):
    """Specify a column by name or index."""

    def __init__(self, columns=None, default=None, help=None):
        # ToDo: make sure we can select by integer
        types = (String,
                 Int)
        self.columns = columns
        super(ColumnLabel, self).__init__(*types, default=default, help=help)

    def validate(self, value):
        """If we are given a column list, make sure that the column provided is valid."""
        super(ColumnLabel, self).validate(value)

        if self.columns:
            if type(value) in bokeh_integer_types:
                if len(self.columns) > value:
                    return
                else:
                    raise ValueError("Not a valid column selection.")
            else:
                if value not in self.columns:
                    raise ValueError("Column provided is not in the list of valid columns: %s" % self.columns)

    def __str__(self):
        return "Column Name or Column String"


class Dimension(HasProps):
    """Configures valid Chart column selections.

    A dimension is Chart property that is assigned one or more columns names or indices. Each
    column can match one or more column types, which are important to charts,
    because the type of column selection can greatly affect the behavior of generalized
    Charts.

    The Dimension also provides convenient utilities for accessing information
    about the current provided configuration at the global, non-grouped level.

    """

    name = String()
    alt_names = Either(String, List(String), default=None)
    columns = Either(ColumnLabel, List(ColumnLabel), default=None)

    valid = Either(PrimitiveProperty, List(PrimitiveProperty), default=None)
    invalid = Either(PrimitiveProperty, List(PrimitiveProperty), default=None)

    selection = Either(ColumnLabel, List(ColumnLabel), default=None)

    def __init__(self, name, **properties):
        properties['name'] = name
        super(Dimension, self).__init__(**properties)
        self._data = pd.DataFrame()
        self._chart_source = None

    def get_valid_types(self, col_data):
        """Returns all property types that are matched."""
        valid_types = list(self.valid)
        matches = []

        # validate each type on the provided column
        for valid_type in valid_types:
            prop = valid_type()

            # if valid, append to the output
            try:
                prop.validate(col_data)
                matches.append(valid_type)
            except ValueError:
                pass

        return matches

    @property
    def data(self):
        """The data selected for the Dimension.

        Returns pd.Series(1) if data is empty or no selection.
        """
        if self._data.empty or self.selection is None:
            return pd.Series(1)
        else:
            # return special column type if available
            if self.selection in list(special_columns.keys()):
                return special_columns[self.selection](self._data)

            return self._data[self.selection]

    def set_data(self, data):
        """Builder must provide data so that builder has access to configuration metadata."""
        self.selection = data[self.name]
        self._chart_source = data
        self._data = data.df
        self.columns = list(self._data.columns.values)

    @property
    def min(self):
        """The minimum of one to many column selections."""
        if isinstance(self.data, pd.Series):
            return self.data.min()
        else:
            return self.data.min(axis=1).min()

    @property
    def max(self):
        """The maximum of one to many column selections."""
        if isinstance(self.data, pd.Series):
            return self.data.max()
        else:
            return self.data.max(axis=1).max()

    @property
    def dtype(self):
        if isinstance(self.data, pd.DataFrame):
            return self.data.dtypes[self.selection[0]]
        else:
            return self.data.dtype

    @property
    def computed(self):
        if self._chart_source is None:
            return False
        else:
            return self._chart_source.is_computed(self.selection)

    @property
    def selected_title(self):
        """A title formatted representation of selected columns."""
        return title_from_columns(self.selection)


class EitherColumn(Either):
    """Allow providing option of column types."""

    # ToDo: incorporate fix into Either
    def matches(self, new, old):
        comparison = super(EitherColumn, self).matches(new, old)
        if isinstance(comparison, bool):
            return comparison
        elif isinstance(comparison, pd.Series):
            return comparison.all()
        else:
            raise ValueError('Failed when comparing Columns')


import numpy as np
import pandas as pd

from bokeh import mpl
from bokeh.plotting import output_file, show

ts = pd.Series(np.random.randn(1000),
               index=pd.date_range('1/1/2000', periods=1000))
ts = ts.cumsum()

p = ts.plot()

output_file("series.html")

show(mpl.to_bokeh())



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2015, Continuum Analytics, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

class AbstractModelStorage(object):
    """ Abstract base class for concrete model storage interfaces.

    """

    def get(self, key):
        """ Returns a JSON object for a given key.

        Args:
            key (str) : key to look up

        Returns:
            data : JSON data (or None, if key does not exist)

        """
        raise NotImplementedError

    def set(self, key, val):
        """ Save a JSON object for a given key.

        Args:
            key (str) : key to set data for
            val (obj) : model data to set

        Returns:
            None

        """
        raise NotImplementedError

    def create(self, key, val):
        """ Save a JSON object for a new key.

        Args:
            key (str) : key to create and set data for
            val (obj) : model data to set

        Returns:
            None

        Raises:
            DataIntegrityException : if the key already exists

        """
        raise NotImplementedError



import seaborn as sns

from bokeh import mpl
from bokeh.plotting import output_file, show

sns.set(style="darkgrid")

# Load the long-form example gammas dataset
gammas = sns.load_dataset("gammas")

# Plot the response with standard error
sns.tsplot(data=gammas, time="timepoint", unit="subject",
           condition="ROI", value="BOLD signal")

output_file("tsplot.html")

show(mpl.to_bokeh())




"""This is the Bokeh charts testing interface.

"""
#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2014, Continuum Analytics, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

from __future__ import absolute_import

import unittest

from bokeh.charts.utils import chunk

#-----------------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------------

class TestBuilder(unittest.TestCase):

    def test_chunk(self):
        chunk_list = list(chunk(range(5), 2))
        self.assertEqual(len(chunk_list), 3)
        self.assertEqual(len(chunk_list[0]), 2)

    # TODO (bev): fix test properly
    # def test_make_scatter(self):
    #     source = ColumnDataSource({"a": [2, 4, 5]})
    #     renderer = make_scatter(source, [0], [1], "circle", "black")
    #     scatter = renderer.glyph
    #     self.assertIsInstance(renderer, GlyphRenderer)
    #     self.assertEqual(renderer.data_source, source)
    #     self.assertEqual(scatter.x, [0])
    #     self.assertEqual(scatter.y, [1])
    #     self.assertIsInstance(scatter, Circle)
    #     self.assertEqual(scatter.line_color, "black")



# The plot server must be running
# Go to http://localhost:5006/bokeh to view this plot

from collections import OrderedDict

import numpy as np
import pandas as pd

from bokeh.plotting import figure, show, output_server
from bokeh.palettes import brewer

N = 20
categories = ['y' + str(x) for x in range(10)]
data = {}
data['x'] = np.arange(N)
for cat in categories:
    data[cat] = np.random.randint(10, 100, size=N)

df = pd.DataFrame(data)
df = df.set_index(['x'])

def stacked(df, categories):
    areas = OrderedDict()
    last = np.zeros(len(df[categories[0]]))
    for cat in categories:
        next = last + df[cat]
        areas[cat] = np.hstack((last[::-1], next))
        last = next
    return areas

areas = stacked(df, categories)

colors = brewer["Spectral"][len(areas)]

x2 = np.hstack((data['x'][::-1], data['x']))

output_server("brewer")

p = figure()
p.patches([x2 for a in areas], list(areas.values()), color=colors, alpha=0.8, line_color=None)

show(p)



from ggplot import aes, geom_line, ggplot, meat
import matplotlib.pyplot as plt

from bokeh import mpl
from bokeh.plotting import output_file, show

g = ggplot(aes(x='date', y='beef'), data=meat) + geom_line()
g.draw()

plt.title("Line ggplot-based plot in Bokeh.")

output_file("line.html")

show(mpl.to_bokeh())



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2015, Continuum Analytics, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import logging
logger = logging.getLogger(__name__)

from collections import defaultdict
import json

from bokeh.exceptions import DataIntegrityException
from bokeh.util.string import decode_utf8

from .abstract_model_storage import AbstractModelStorage
from .backbone_storage import BackboneStorage

class InMemoryModelStorage(AbstractModelStorage):
    """ Concrete server model storage using in-memory data structures for
    storage.

    .. warning::
        This storage backend is not persistent.

    """
    def __init__(self):
        self._data = {}

    def get(self, key):
        data = self._data.get(key, None)
        if data is None:
            return None
        attrs = json.loads(decode_utf8(data))
        return attrs

    def set(self, key, val):
        self._data[key] = json.dumps(val)

    def create(self, key, val):
        if key in self._data:
            raise DataIntegrityException("%s already exists" % key)
        self._data[key] = json.dumps(val)

class InMemoryBackboneStorage(BackboneStorage):
    """ Concrete BackBone  storage using in-memory data structures for
    storage.

    .. warning::
        This storage backend is not persistent.

    """
    def __init__(self):
        self._inmem_data = {}
        self._inmem_sets = defaultdict(set)

    def mget(self, doc_keys):
        return [self._inmem_data.get(key, None) for key in doc_keys]

    def mset(self, data):
        self._inmem_data.update(data)

    def sadd(self, doc_key, *keys):
        self._inmem_sets[doc_key].update(keys)

    def srem(self, doc_key, member_key):
        inmem_set = self._inmem_sets[doc_key]
        try: inmem_set.remove(member_key)
        except KeyError: pass

    def smembers(self, doc_key):
        return list(self._inmem_sets[doc_key])

    def set(self, key, data):
        self._inmem_data[key] = data

    def delete(self, key):
        del self._inmem_data[key]




from __future__ import absolute_import

import unittest

from bokeh.plotting import figure
from bokeh.models.ranges import DataRange1d


class TestGlyphRenderer(unittest.TestCase):
    def test_warning_about_colons_in_column_labels_for_axis(self):
        invalid_labels = ['0', '1', '2:0']
        plot = figure(
            x_range=invalid_labels,
            y_range=invalid_labels,
            plot_width=900,
            plot_height=400,
        )

        errors = plot._check_colon_in_category_label()

        self.assertEqual(errors, [(
            1003,
            'MALFORMED_CATEGORY_LABEL',
            'Category labels cannot contain colons',
            '[range:x_range] [first_value: 2:0] '
            '[range:y_range] [first_value: 2:0] '
            '[renderer: Figure, ViewModel:Plot, ref _id: '
            '%s]' % plot._id
        )])

    def test_validates_colons_only_in_factorial_range(self):
        plot = figure(
            x_range=DataRange1d(start=0.0, end=2.2),
            y_range=['0', '1', '2:0'],
            plot_width=900,
            plot_height=400,
        )

        errors = plot._check_colon_in_category_label()

        self.assertEqual(errors, [(
            1003,
            'MALFORMED_CATEGORY_LABEL',
            'Category labels cannot contain colons',
            '[range:y_range] [first_value: 2:0] '
            '[renderer: Figure, ViewModel:Plot, ref _id: '
            '%s]' % plot._id
        )])




if __name__ == '__main__':
    unittest.main()



"""This is the Bokeh charts interface. It gives you a high level API to build
complex plot is a simple way.

This is the ChartObject class, a minimal prototype class to build more chart
types on top of it. It provides the mechanisms to support the shared chained
methods.
"""
#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2014, Continuum Analytics, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

from __future__ import absolute_import

from six import string_types
from collections import OrderedDict
from ..properties import bokeh_integer_types, Datetime

try:
    import numpy as np

except ImportError:
    np = None

try:
    import pandas as pd

except ImportError:
    pd = None
try:
    import blaze
except ImportError:
    blaze=None
#-----------------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------------

DEFAULT_INDEX_ALIASES = list('abcdefghijklmnopqrstuvz1234567890')
DEFAULT_INDEX_ALIASES += list(zip(DEFAULT_INDEX_ALIASES, DEFAULT_INDEX_ALIASES))

class DataAdapter(object):
    """
    Adapter object used to normalize Charts inputs to a common interface.
    Supported inputs are dict, list, tuple, np.ndarray and pd.DataFrame.
    """
    def __init__(self, data, index=None, columns=None, force_alias=True):
        self.__values = data
        self._values = self.validate_values(data)

        self.convert_index_to_int = False
        self._columns_map = {}
        self.convert_items_to_dict = False

        if columns is None and force_alias:
            # no column 'labels' defined for data... in this case we use
            # default names
            keys = getattr(self._values, 'keys', None)
            if callable(keys):
                columns = list(keys())
            elif keys is None:
                columns = list(map(str, range(len(data))))
            else:
                columns = list(keys)

        if columns:
            self._columns = columns

            # define a mapping between the real keys to access data and the aliases
            # we have defined using 'columns'
            self._columns_map = dict(zip(columns, self.keys()))

        if index is not None:
            self._index = index
            self.convert_items_to_dict = True

        elif force_alias:
            _index = getattr(self._values, 'index', None)

            # check because if it is a callable self._values is not a
            # dataframe (probably a list)
            if _index is None:
                indexes = self.index

                if isinstance(indexes[0], int):
                    self._index = DEFAULT_INDEX_ALIASES[:][:len(self.values()[0])]
                    self.convert_items_to_dict = True

            elif not callable(_index):
                self._index = list(_index)
                self.convert_items_to_dict = True

            else:
                self._index = DEFAULT_INDEX_ALIASES[:][:len(self.values()[0])]
                self.convert_items_to_dict = True

    @staticmethod
    def is_number(value):
        numbers = (float, ) + bokeh_integer_types
        return isinstance(value, numbers)

    @staticmethod
    def is_datetime(value):
        try:
            dt = Datetime(value)
            dt # shut up pyflakes
            return True

        except ValueError:
            return False

    @staticmethod
    def validate_values(values):
        if np and isinstance(values, np.ndarray):
            if len(values.shape) == 1:
                return np.array([values])
            else:
                return values

        elif pd and isinstance(values, pd.DataFrame):
            return values

        elif isinstance(values, (dict, OrderedDict)):
            if all(DataAdapter.is_number(x) for x in values.values()):
                return values

            return values

        elif isinstance(values, (list, tuple)):
            if all(DataAdapter.is_number(x) for x in values):
                return [values]

            return values

        elif hasattr(values, '__array__'):
            values = pd.DataFrame(np.asarray(values))
            return values

        # TODO: Improve this error message..
        raise TypeError("Input type not supported! %s" % values)


    def index_converter(self, x):
        key = self._columns_map.get(x, x)
        if self.convert_index_to_int:
            key = int(key)
        return key

    def keys(self):
        # assuming it's a dict or dataframe
        keys = getattr(self._values, "keys", None)

        if callable(keys):
            return list(keys())

        elif keys is None:
            self.convert_index_to_int = True
            indexes = range(len(self._values))
            return list(map(str, indexes))

        else:
            return list(keys)

    def __len__(self):
        return len(self.values())

    def __iter__(self):
        for k in self.keys():
            yield k

    def __getitem__(self, key):
        val = self._values[self.index_converter(key)]

        # if we have "index aliases" we need to remap the values...
        if self.convert_items_to_dict:
            val = dict(zip(self._index, val))

        return val

    def values(self):
        return self.normalize_values(self._values)

    @staticmethod
    def normalize_values(values):
        _values = getattr(values, "values", None)

        if callable(_values):
            return list(_values())

        elif _values is None:
            return values

        else:
            # assuming it's a dataframe, in that case it returns transposed
            # values compared to it's dict equivalent..
            return list(_values.T)

    def items(self):
        return [(key, self[key]) for key in self]

    def iterkeys(self):
        return iter(self)

    def itervalues(self):
        for k in self:
            yield self[k]

    def iteritems(self):
        for k in self:
            yield (k, self[k])

    @property
    def columns(self):
        try:
            return self._columns

        except AttributeError:
            return list(self.keys())

    @property
    def index(self):
        try:
            return self._index

        except AttributeError:
            index = getattr(self._values, "index", None)

            if not callable(index) and index is not None:
                # guess it's a pandas dataframe..
                return index

        # no, it's not. So it's probably a list so let's get the
        # values and check
        values = self.values()

        if isinstance(values, dict):
            return list(values.keys())

        else:
            first_el = self.values()[0]

            if isinstance(first_el, dict):
                indexes = list(first_el.keys())

            else:
                indexes = range(0, len(self.values()[0]))
                self._index = indexes
            return indexes

    #-----------------------------------------------------------------------------
    # Convenience methods
    #-----------------------------------------------------------------------------
    @staticmethod
    def get_index_and_data(values, index=None):
        """Parse values (that must be one of the DataAdapter supported
        input types) and create an separate/create index and data
        depending on values type and index.

        Args:
            values (iterable): container that holds data to be plotted using
                on the Chart classes

        Returns:
            A tuple of (index, values), where: ``index`` is an iterable that
            represents the data index and ``values`` is an iterable containing
            the values to be plotted.

        """
        _values = DataAdapter(values, force_alias=False)
        if hasattr(values, 'keys'):
            if index is not None:
                if isinstance(index, string_types):
                    xs = _values[index]
                else:
                    xs = index
            else:
                try:
                    xs = _values.index
                except AttributeError:
                    xs = values.index
        else:
            if index is None:
                xs = _values.index
            elif isinstance(index, string_types):
                xs = _values[index]
            else:
                xs = index

        return xs, _values



""" Functions useful for string manipulations or encoding.

"""
from __future__ import absolute_import

def encode_utf8(u):
    """ Encode a UTF-8 string to a sequence of bytes.

    Args:
        u (str) : the string to encode

    Returns:
        bytes

    """
    import sys
    if sys.version_info[0] == 2:
        u = u.encode('utf-8')
    return u

def decode_utf8(u):
    """ Decode a sequence of bytes to a UTF-8 string

    Args:
        u (str) : the bytes to decode

    Returns:
        UTF-8 string

    """
    import sys
    if sys.version_info[0] == 2:
        u = u.decode('utf-8')
    return u

def nice_join(seq, sep=", "):
    """ Join together sequences of strings into English-friendly phrases using
    the conjunction ``or`` when appropriate.

    Args:
        seq (seq[str]) : a sequence of strings to nicely join
        sep (str, optional) : a sequence delimiter to use (default: ", ")

    Returns:
        a joined string

    Examples:
        >>> nice_join(["a", "b", "c"])
        'a, b or c'

    """
    seq = [str(x) for x in seq]

    if len(seq) <= 1:
        return sep.join(seq)
    else:
        return "%s or %s" % (sep.join(seq[:-1]), seq[-1])


from collections import OrderedDict

from bokeh._legacy_charts import Step, show, output_file

xyvalues = OrderedDict(
    python=[2, 3, 7, 5, 26, 81, 44, 93, 94, 105, 66, 67, 90, 83],
    pypy=[12, 20, 47, 15, 126, 121, 144, 333, 354, 225, 276, 287, 270, 230],
    jython=[22, 43, 70, 75, 76, 101, 114, 123, 194, 215, 201, 227, 139, 160],
)

# any of the following commented are also valid Step inputs
#xyvalues = pd.DataFrame(xyvalues)
#xyvalues = list(xyvalues.values())
#xyvalues = np.array(list(xyvalues.values()))

output_file("steps.html", title="line.py example")

chart = Step(xyvalues, title="Steps", ylabel='measures', legend='top_left')

show(chart)



__icons__ = [
    "adjust", "adn", "align-center", "align-justify", "align-left", "align-right", "ambulance",
    "anchor", "android", "angellist", "angle-double-down", "angle-double-left", "angle-double-right",
    "angle-double-up", "angle-down", "angle-left", "angle-right", "angle-up", "apple", "archive",
    "area-chart", "arrow-circle-down", "arrow-circle-left", "arrow-circle-o-down", "arrow-circle-o-left",
    "arrow-circle-o-right", "arrow-circle-o-up", "arrow-circle-right", "arrow-circle-up", "arrow-down",
    "arrow-left", "arrow-right", "arrow-up", "arrows", "arrows-alt", "arrows-h", "arrows-v", "asterisk",
    "at", "automobile", "backward", "ban", "bank", "bar-chart", "bar-chart-o", "barcode", "bars", "beer",
    "behance", "behance-square", "bell", "bell-o", "bell-slash", "bell-slash-o", "bicycle", "binoculars",
    "birthday-cake", "bitbucket", "bitbucket-square", "bitcoin", "bold", "bolt", "bomb", "book", "bookmark",
    "bookmark-o", "briefcase", "btc", "bug", "building", "building-o", "bullhorn", "bullseye", "bus", "cab",
    "calculator", "calendar", "calendar-o", "camera", "camera-retro", "car", "caret-down", "caret-left",
    "caret-right", "caret-square-o-down", "caret-square-o-left", "caret-square-o-right", "caret-square-o-up",
    "caret-up", "cc", "cc-amex", "cc-discover", "cc-mastercard", "cc-paypal", "cc-stripe", "cc-visa", "certificate",
    "chain", "chain-broken", "check", "check-circle", "check-circle-o", "check-square", "check-square-o",
    "chevron-circle-down", "chevron-circle-left", "chevron-circle-right", "chevron-circle-up", "chevron-down",
    "chevron-left", "chevron-right", "chevron-up", "child", "circle", "circle-o", "circle-o-notch", "circle-thin",
    "clipboard", "clock-o", "close", "cloud", "cloud-download", "cloud-upload", "cny", "code", "code-fork",
    "codepen", "coffee", "cog", "cogs", "columns", "comment", "comment-o", "comments", "comments-o", "compass",
    "compress", "copy", "copyright", "credit-card", "crop", "crosshairs", "css3", "cube", "cubes", "cut", "cutlery",
    "dashboard", "database", "dedent", "delicious", "desktop", "deviantart", "digg", "dollar", "dot-circle-o",
    "download", "dribbble", "dropbox", "drupal", "edit", "eject", "ellipsis-h", "ellipsis-v", "empire", "envelope",
    "envelope-o", "envelope-square", "eraser", "eur", "euro", "exchange", "exclamation", "exclamation-circle",
    "exclamation-triangle", "expand", "external-link", "external-link-square", "eye", "eye-slash", "eyedropper",
    "facebook", "facebook-square", "fast-backward", "fast-forward", "fax", "female", "fighter-jet", "file",
    "file-archive-o", "file-audio-o", "file-code-o", "file-excel-o", "file-image-o", "file-movie-o", "file-o",
    "file-pdf-o", "file-photo-o", "file-picture-o", "file-powerpoint-o", "file-sound-o", "file-text", "file-text-o",
    "file-video-o", "file-word-o", "file-zip-o", "files-o", "film", "filter", "fire", "fire-extinguisher", "flag",
    "flag-checkered", "flag-o", "flash", "flask", "flickr", "floppy-o", "folder", "folder-o", "folder-open",
    "folder-open-o", "font", "forward", "foursquare", "frown-o", "futbol-o", "gamepad", "gavel", "gbp", "ge",
    "gear", "gears", "gift", "git", "git-square", "github", "github-alt", "github-square", "gittip", "glass",
    "globe", "google", "google-plus", "google-plus-square", "google-wallet", "graduation-cap", "group",
    "h-square", "hacker-news", "hand-o-down", "hand-o-left", "hand-o-right", "hand-o-up", "hdd-o", "header",
    "headphones", "heart", "heart-o", "history", "home", "hospital-o", "html5", "ils", "image", "inbox", "indent",
    "info", "info-circle", "inr", "instagram", "institution", "ioxhost", "italic", "joomla", "jpy", "jsfiddle",
    "key", "keyboard-o", "krw", "language", "laptop", "lastfm", "lastfm-square", "leaf", "legal", "lemon-o",
    "level-down", "level-up", "life-bouy", "life-buoy", "life-ring", "life-saver", "lightbulb-o", "line-chart",
    "link", "linkedin", "linkedin-square", "linux", "list", "list-alt", "list-ol", "list-ul", "location-arrow",
    "lock", "long-arrow-down", "long-arrow-left", "long-arrow-right", "long-arrow-up", "magic", "magnet", "mail-forward",
    "mail-reply", "mail-reply-all", "male", "map-marker", "maxcdn", "meanpath", "medkit", "meh-o", "microphone",
    "microphone-slash", "minus", "minus-circle", "minus-square", "minus-square-o", "mobile", "mobile-phone", "money",
    "moon-o", "mortar-board", "music", "navicon", "newspaper-o", "openid", "outdent", "pagelines", "paint-brush",
    "paper-plane", "paper-plane-o", "paperclip", "paragraph", "paste", "pause", "paw", "paypal", "pencil", "pencil-square",
    "pencil-square-o", "phone", "phone-square", "photo", "picture-o", "pie-chart", "pied-piper", "pied-piper-alt",
    "pinterest", "pinterest-square", "plane", "play", "play-circle", "play-circle-o", "plug", "plus", "plus-circle",
    "plus-square", "plus-square-o", "power-off", "print", "puzzle-piece", "qq", "qrcode", "question", "question-circle",
    "quote-left", "quote-right", "ra", "random", "rebel", "recycle", "reddit", "reddit-square", "refresh", "remove",
    "renren", "reorder", "repeat", "reply", "reply-all", "retweet", "rmb", "road", "rocket", "rotate-left", "rotate-right",
    "rouble", "rss", "rss-square", "rub", "ruble", "rupee", "save", "scissors", "search", "search-minus", "search-plus",
    "send", "send-o", "share", "share-alt", "share-alt-square", "share-square", "share-square-o", "shekel", "sheqel",
    "shield", "shopping-cart", "sign-in", "sign-out", "signal", "sitemap", "skype", "slack", "sliders", "slideshare", "smile-o",
    "soccer-ball-o", "sort", "sort-alpha-asc", "sort-alpha-desc", "sort-amount-asc", "sort-amount-desc", "sort-asc",
    "sort-desc", "sort-down", "sort-numeric-asc", "sort-numeric-desc", "sort-up", "soundcloud", "space-shuttle", "spinner",
    "spoon", "spotify", "square", "square-o", "stack-exchange", "stack-overflow", "star", "star-half", "star-half-empty",
    "star-half-full", "star-half-o", "star-o", "steam", "steam-square", "step-backward", "step-forward", "stethoscope",
    "stop", "strikethrough", "stumbleupon", "stumbleupon-circle", "subscript", "suitcase", "sun-o", "superscript",
    "support", "table", "tablet", "tachometer", "tag", "tags", "tasks", "taxi", "tencent-weibo", "terminal", "text-height",
    "text-width", "th", "th-large", "th-list", "thumb-tack", "thumbs-down", "thumbs-o-down", "thumbs-o-up", "thumbs-up",
    "ticket", "times", "times-circle", "times-circle-o", "tint", "toggle-down", "toggle-left", "toggle-off", "toggle-on",
    "toggle-right", "toggle-up", "trash", "trash-o", "tree", "trello", "trophy", "truck", "try", "tty", "tumblr",
    "tumblr-square", "turkish-lira", "twitch", "twitter", "twitter-square", "umbrella", "underline", "undo", "university",
    "unlink", "unlock", "unlock-alt", "unsorted", "upload", "usd", "user", "user-md", "users", "video-camera", "vimeo-square",
    "vine", "vk", "volume-down", "volume-off", "volume-up", "warning", "wechat", "weibo", "weixin", "wheelchair", "wifi",
    "windows", "won", "wordpress", "wrench", "xing", "xing-square", "yahoo", "yelp", "yen", "youtube", "youtube-play",
    "youtube-square",
]



""" Helpers for testing builders

"""
#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2014, Continuum Analytics, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------


#-----------------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------------


def create_chart(klass, values, compute_values=True, **kws):
    """ Create a new chart klass instance with values and the extra kws keyword
    parameters.

    Args:
        klass (class): chart class to be created
        values (iterable): chart data series
        compute_values (bool): if == True underlying chart attributes (like data,
                ranges, source, etc..) are computed by calling _setup_show,
                _prepare_show and _show_teardown methods.
        **kws (refer to klass arguments specification details)

    Return:
        _chart: klass chart instance
    """
    _chart = klass(
        values, title="title", xlabel="xlabel", ylabel="ylabel",
        legend="top_left", xscale="linear", yscale="linear",
        width=800, height=600, tools=True,
        filename=False, server=False, notebook=False,
        **kws
    )

    return _chart



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2015, Continuum Analytics, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
from __future__ import absolute_import

from bokeh.exceptions import UnauthorizedException

from bokeh.server.app import bokeh_app
from bokeh.server.models import user
from bokeh.util.testing import BokehServerTestCase


class TestUser(BokehServerTestCase):
    def setUp(self):
        super(TestUser, self).setUp()
        self.client = bokeh_app.servermodel_storage

    def test_cant_create_twice(self):
        user.new_user(self.client, 'test@test.com', 'mypassword', docs=[1, 2, 3])
        self.assertRaises(UnauthorizedException, user.new_user,
                          self.client, 'test@test.com', 'mypassword')

    def test_auth_user(self):
        self.assertRaises(UnauthorizedException,
                          user.auth_user,
                          self.client, 'test@test.com', 'mypassword')
        model = user.new_user(self.client, 'test@test.com', 'mypassword')
        assert model.username == 'test@test.com'
        model = user.auth_user(self.client, 'test@test.com', 'mypassword')
        self.assertRaises(UnauthorizedException, user.auth_user,
                          self.client, 'test@test.com', 'wrongpassword')



# The plot server must be running
# Go to http://localhost:5006/bokeh to view this plot

from __future__ import division

import numpy as np

from bokeh.plotting import figure, show, output_server, vplot

def streamlines(x, y, u, v, density=1):
    '''Returns streamlines of a vector flow.

    * x and y are 1d arrays defining an *evenly spaced* grid.
    * u and v are 2d arrays (shape [y,x]) giving velocities.
    * density controls the closeness of the streamlines. For different
      densities in each direction, use a tuple or list [densityx, densityy].

     '''

    ## Set up some constants - size of the grid used.
    NGX = len(x)
    NGY = len(y)

    ## Constants used to convert between grid index coords and user coords.
    DX = x[1]-x[0]
    DY = y[1]-y[0]
    XOFF = x[0]
    YOFF = y[0]

    ## Now rescale velocity onto axes-coordinates
    u = u / (x[-1]-x[0])
    v = v / (y[-1]-y[0])
    speed = np.sqrt(u*u+v*v)
    ## s (path length) will now be in axes-coordinates, but we must
    ## rescale u for integrations.
    u *= NGX
    v *= NGY
    ## Now u and v in grid-coordinates.

    NBX = int(30*density)
    NBY = int(30*density)
    blank = np.zeros((NBY,NBX))

    bx_spacing = NGX/float(NBX-1)
    by_spacing = NGY/float(NBY-1)

    def blank_pos(xi, yi):
        return int((xi / bx_spacing) + 0.5), \
               int((yi / by_spacing) + 0.5)

    def value_at(a, xi, yi):
        if type(xi) == np.ndarray:
            x = xi.astype(np.int)
            y = yi.astype(np.int)
        else:
            x = np.int(xi)
            y = np.int(yi)
        a00 = a[y,x]
        a01 = a[y,x+1]
        a10 = a[y+1,x]
        a11 = a[y+1,x+1]
        xt = xi - x
        yt = yi - y
        a0 = a00*(1-xt) + a01*xt
        a1 = a10*(1-xt) + a11*xt
        return a0*(1-yt) + a1*yt

    def rk4_integrate(x0, y0):
        ## This function does RK4 forward and back trajectories from
        ## the initial conditions, with the odd 'blank array'
        ## termination conditions. TODO tidy the integration loops.

        def f(xi, yi):
            dt_ds = 1./value_at(speed, xi, yi)
            ui = value_at(u, xi, yi)
            vi = value_at(v, xi, yi)
            return ui*dt_ds, vi*dt_ds

        def g(xi, yi):
            dt_ds = 1./value_at(speed, xi, yi)
            ui = value_at(u, xi, yi)
            vi = value_at(v, xi, yi)
            return -ui*dt_ds, -vi*dt_ds

        check = lambda xi, yi: xi>=0 and xi<NGX-1 and yi>=0 and yi<NGY-1

        bx_changes = []
        by_changes = []

        ## Integrator function
        def rk4(x0, y0, f):
            ds = 0.01 #min(1./NGX, 1./NGY, 0.01)
            stotal = 0
            xi = x0
            yi = y0
            xb, yb = blank_pos(xi, yi)
            xf_traj = []
            yf_traj = []
            while check(xi, yi):
                # Time step. First save the point.
                xf_traj.append(xi)
                yf_traj.append(yi)
                # Next, advance one using RK4
                try:
                    k1x, k1y = f(xi, yi)
                    k2x, k2y = f(xi + .5*ds*k1x, yi + .5*ds*k1y)
                    k3x, k3y = f(xi + .5*ds*k2x, yi + .5*ds*k2y)
                    k4x, k4y = f(xi + ds*k3x, yi + ds*k3y)
                except IndexError:
                    # Out of the domain on one of the intermediate steps
                    break
                xi += ds*(k1x+2*k2x+2*k3x+k4x) / 6.
                yi += ds*(k1y+2*k2y+2*k3y+k4y) / 6.
                # Final position might be out of the domain
                if not check(xi, yi): break
                stotal += ds
                # Next, if s gets to thres, check blank.
                new_xb, new_yb = blank_pos(xi, yi)
                if new_xb != xb or new_yb != yb:
                    # New square, so check and colour. Quit if required.
                    if blank[new_yb,new_xb] == 0:
                        blank[new_yb,new_xb] = 1
                        bx_changes.append(new_xb)
                        by_changes.append(new_yb)
                        xb = new_xb
                        yb = new_yb
                    else:
                        break
                if stotal > 2:
                    break
            return stotal, xf_traj, yf_traj

        integrator = rk4

        sf, xf_traj, yf_traj = integrator(x0, y0, f)
        sb, xb_traj, yb_traj = integrator(x0, y0, g)
        stotal = sf + sb
        x_traj = xb_traj[::-1] + xf_traj[1:]
        y_traj = yb_traj[::-1] + yf_traj[1:]

        ## Tests to check length of traj. Remember, s in units of axes.
        if len(x_traj) < 1: return None
        if stotal > .2:
            initxb, inityb = blank_pos(x0, y0)
            blank[inityb, initxb] = 1
            return x_traj, y_traj
        else:
            for xb, yb in zip(bx_changes, by_changes):
                blank[yb, xb] = 0
            return None

    ## A quick function for integrating trajectories if blank==0.
    trajectories = []
    def traj(xb, yb):
        if xb < 0 or xb >= NBX or yb < 0 or yb >= NBY:
            return
        if blank[yb, xb] == 0:
            t = rk4_integrate(xb*bx_spacing, yb*by_spacing)
            if t != None:
                trajectories.append(t)

    ## Now we build up the trajectory set. I've found it best to look
    ## for blank==0 along the edges first, and work inwards.
    for indent in range((max(NBX,NBY))//2):
        for xi in range(max(NBX,NBY)-2*indent):
            traj(xi+indent, indent)
            traj(xi+indent, NBY-1-indent)
            traj(indent, xi+indent)
            traj(NBX-1-indent, xi+indent)

    xs = [np.array(t[0])*DX+XOFF for t in trajectories]
    ys = [np.array(t[1])*DY+YOFF for t in trajectories]

    return xs, ys

xx = np.linspace(-3, 3, 100)
yy = np.linspace(-3, 3, 100)

Y, X = np.meshgrid(xx, yy)
U = -1 - X**2 + Y
V = 1 + X - Y**2
speed = np.sqrt(U*U + V*V)
theta = np.arctan(V/U)

x0 = X[::2, ::2].flatten()
y0 = Y[::2, ::2].flatten()
length = speed[::2, ::2].flatten()/40
angle = theta[::2, ::2].flatten()
x1 = x0 + length * np.cos(angle)
y1 = y0 + length * np.sin(angle)

xs, ys = streamlines(xx, yy, U.T, V.T, density=2)

cm = np.array(["#C7E9B4", "#7FCDBB", "#41B6C4", "#1D91C0", "#225EA8", "#0C2C84"])
ix = ((length-length.min())/(length.max()-length.min())*5).astype('int')
colors = cm[ix]

output_server("vector")

p1 = figure()
p1.segment(x0, y0, x1, y1, color=colors, line_width=2)

p2 = figure()
p2.multi_line(xs, ys, color="#ee6666", line_width=2, line_alpha=0.8)

show(vplot(p1,p2))  # open a browser



''' The templates module contains templates used by Bokeh to
enable embedding Bokeh plots in various ways.

Attributes:
    JS_RESOURCES: This template is for loading BokehJS code and css
        according to the configuration in a Resources object.

Args:
    js_files (list[str]) : a list of URIs for JS files to include
    js_raw (list[str]) : a list of raw JS snippets to put between `<style>` tags

Attributes:
    CSS_RESOURCES: This template is for loading Bokeh css
        according to the configuration in a Resources object.

Args:
    css_files (list[str]) : a list of URIs for CSS files to include
    css_raw (list[str]) : a list of raw CSS snippets to put between `<style>` tags

Attributes:
    PLOT_DIV: This template is for creating a basic plot div (to be
        used in conjunction with PLOT_JS).

Args:
    elementid (str) : a unique identifier for the div
        a PLOT_JS template should be configured with the same elementid

Attributes:
    PLOT_JS: This template is for creating the JavaScript code snippet
        that can render a plot into a corresponding PLOT_DIV.

Args:
    modelid (str) : The Bokeh model id for the object to render
        typically for a Plot, PlotContext, etc.
    modeltype (str) : the type of the model to render
        used to reference the appropriate Backbone collection
    elementid (str) : the id of the div to render the plot into

Attributes:
    PLOT_SCRIPT: This template is for creating a full ``<script>`` tag
        for raw JS code. Useful with the PLOT_JS template.

Args:
    plot_js (str) : raw JavaScript code to include

Attributes:
    FILE: This template is for rendering Bokeh plots into a basic .html file.

Args:
    title (str) : a value for `<title>` tags
    plot_resources (str) : typically the output of RESOURCES
    plot_script (str) : typically the output of PLOT_SCRIPT
    plot_div (str) : typically the output of PLOT_DIV

    Users can customize the file output by providing their own template
    with these parameters.

Attributes:
    NOTEBOOK_LOAD: This template is for loading BokehJS code and CSS
        into the IPython Notebook according to a resources configuration.

Args:
    plot_resources (str) : typically the output of RESOURCES
    log_url (str) : URL to Bokeh logo to dispay
    verbose (bool) : whether to display verbose info about BokehJS configuration, etc
    bokeh_version (str) : the current version of Bokeh
    js_info (str) : information about the location, version, etc. of BokehJS code
    css_info (str) : information about the location, version, etc. of BokehJS css
    warnings (list[str]) : list of warnings to display to user

Attributes:
    NOTEBOOK_DIV: This template is for rendering a Bokeh plot into the
        IPython Notebook.

Args:
    plot_script (str) : typically the output of PLOT_SCRIPT
    plot_div (str) : typically the output of PLOT_DIV

Attributes:
    AUTOLOAD: This template is for creating an "autoload" JS script.
        The script automatically and asynchronously loads BokehJS (if
        necessary) and then replaces any suitably constructed
        ``<script>`` tag that loads it with the rendered plot.

Attributes:
    AUTOLOAD_SERVER: This template is for creating ``<script>`` tags
        that run AUTOLOAD scripts for plots that connect to a Bokeh Server
        for their data

Args:
    src_path (str) : path to AUTOLOAD script
    elementid (str) : the a unique id for the script tag
    modelid (str) : The Bokeh model id for the object to render
        typically for a Plot, PlotContext, etc.
    root_url (str) : root URL of the Bokeh Server
    docid (str) : document ID for the document on the server to load
    docapikey (str) : API key for the document

Attributes:
    AUTOLOAD_STATIC: This template is for creating ``<script>`` tags
        that run AUTOLOAD scripts for plots that have their data embedded
        in the AUTOLOAD script

Args:
    src_path (str) : path to AUTOLOAD script
    elementid (str) : the a unique id for the script tag
    modelid (str) : The Bokeh model id for the object to render
        typically for a Plot, PlotContext, etc.
    modeltype (str) : the type of the model to render
        used to reference the appropriate Backbone collection

.. note:: This script injects a ``<div>`` in place, so must be placed under ``<body>``.

'''

from __future__ import absolute_import

from os.path import abspath, join, split

import jinja2

_templates_path = join(abspath(split(__file__)[0]), "_templates")

JS_RESOURCES = jinja2.Template(
    open(join(_templates_path, "js_resources.html")).read()
)

CSS_RESOURCES = jinja2.Template(
    open(join(_templates_path, "css_resources.html")).read()
)


PLOT_DIV = jinja2.Template(
    open(join(_templates_path, "plot_div.html")).read()
)


PLOT_JS = jinja2.Template(
    open(join(_templates_path, "plot_js.js")).read()
)

PLOT_SCRIPT = jinja2.Template(
    open(join(_templates_path, "plot_script.html")).read()
)


FILE = jinja2.Template(
    open(join(_templates_path, "file.html")).read()
)


NOTEBOOK_LOAD = jinja2.Template(
    open(join(_templates_path, "notebook_load.html")).read()
)


NOTEBOOK_DIV = jinja2.Template(
    open(join(_templates_path, "notebook_div.html")).read()
)


AUTOLOAD = jinja2.Template(
    open(join(_templates_path, "autoload.js")).read()
)


AUTOLOAD_SERVER = jinja2.Template(
    open(join(_templates_path, "autoload_server.html")).read()
)

AUTOLOAD_SERVER_PUBLIC = jinja2.Template(
    open(join(_templates_path, "autoload_server.html")).read()
)


AUTOLOAD_STATIC= jinja2.Template(
    open(join(_templates_path, "autoload_static.html")).read()
)



""" Models for describing different kinds of ranges of values
in different kinds of spaces (e.g., continuous or categorical)
and with options for "auto sizing".

"""
from __future__ import absolute_import

from ..plot_object import PlotObject
from ..properties import abstract
from ..properties import Int, Float, String, Datetime, Instance, List, Either
from .callbacks import Callback
from .renderers import Renderer

@abstract
class Range(PlotObject):
    """ A base class for all range types. ``Range`` is not generally
    useful to instantiate on its own.

    """

    callback = Instance(Callback, help="""
    A callback to run in the browser whenever the range is updated.
    """)

class Range1d(Range):
    """ A fixed, closed range [start, end] in a continuous scalar
    dimension.

    In addition to supplying ``start`` and ``end`` keyword arguments
    to the ``Range1d`` initializer, you can also instantiate with
    the convenience syntax::

        Range(0, 10) # equivalent to Range(start=0, end=10)

    """

    start = Either(Float, Datetime, Int, help="""
    The start of the range.
    """)

    end = Either(Float, Datetime, Int, help="""
    The end of the range.
    """)

    def __init__(self, *args, **kwargs):
        if args and ('start' in kwargs or 'end' in kwargs):
            raise ValueError("'start' and 'end' keywords cannot be used with positional arguments")
        elif args and len(args) != 2:
            raise ValueError('Only Range1d(start, end) acceptable when using positional arguments')
        elif args:
            kwargs['start'] = args[0]
            kwargs['end'] = args[1]
        super(Range1d, self).__init__(**kwargs)

@abstract
class DataRange(Range):
    """ A base class for all data range types. ``DataRange`` is not
    generally useful to instantiate on its own.

    """

    names = List(String, help="""
    A list of names to query for. If set, only renderers that
    have a matching value for their ``name`` attribute will be used
    for autoranging.
    """)

    renderers = List(Instance(Renderer), help="""
    An explicit list of renderers to autorange against. If unset,
    defaults to all renderers on a plot.
    """)

class DataRange1d(DataRange):
    """ An auto-fitting range in a continuous scalar dimension.

    """

    range_padding = Float(0.1, help="""
    A percentage of the total range size to add as padding to
    the range start and end.
    """)

    start = Float(help="""
    An explicitly supplied range start. If provided, will override
    automatically computed start value.
    """)

    end = Float(help="""
    An explicitly supplied range end. If provided, will override
    automatically computed end value.
    """)

class FactorRange(Range):
    """ A range in a categorical dimension.

    In addition to supplying ``factors`` keyword argument to the
    ``FactorRange`` initializer, you can also instantiate with
    the convenience syntax::

        FactorRange("foo", "bar") # equivalent to FactorRange(factors=["foo", "bar"])

    .. note::
        ``FactorRange`` may be renamed to ``CategoricalRange`` in
        the future.

    """

    offset = Float(0, help="""
    An offset to the (synthetic) range (default: 0)

    .. note::
        The primary usage of this is to support compatibility and integration
        with other plotting systems, and will not generally of interest to
        most users.
    """)

    factors = Either(List(String), List(Int), help="""
    A list of string or integer factors (categories) to comprise
    this categorical range.
    """)

    def __init__(self, *args, **kwargs):
        if args and "factors" in kwargs:
            raise ValueError("'factors' keyword cannot be used with positional arguments")
        elif args:
            kwargs['factors'] = list(args)
        super(FactorRange, self).__init__(**kwargs)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2015, Continuum Analytics, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
from __future__ import absolute_import

import logging
logger = logging.getLogger(__name__)

from functools import wraps

from bokeh.exceptions import AuthenticationException
from flask import abort, jsonify

from ..app import bokeh_app
from ..models import docs


def handle_auth_error(func):
    """Decorator wraps a function and watches for AuthenticationException
    If one is thrown, log and abort 401 instead
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except AuthenticationException as e:
            logger.exception(e)
            return abort(401)
    return wrapper

def check_read_authentication(func):
    @wraps(func)
    def wrapper(docid, *args, **kwargs):
        if bokeh_app.authentication.can_read_doc(docid):
            return func(docid, *args, **kwargs)
        else:
            abort(401)
    return wrapper

def check_write_authentication(func):
    @wraps(func)
    def wrapper(docid, *args, **kwargs):
        if bokeh_app.authentication.can_write_doc(docid):
            return func(docid, *args, **kwargs)
        else:
            abort(401)
    return wrapper

@bokeh_app.route('/bokeh/login', methods=['GET'])
def login_get():
    ''' Log in a user from a form.

    :status 200: render login view

    '''
    return bokeh_app.authentication.login_get()

@bokeh_app.route('/bokeh/login', methods=['POST'])
def login_post():
    ''' Log in user from a submission.

    :status 200: if API flag set, log in status
    :status 302: if API flag not set, redirect to index on
        success, to login on failue

    '''
    return bokeh_app.authentication.login_post()

@bokeh_app.route('/bokeh/loginfromapikey', methods=['GET'])
def login_from_apikey():
    ''' Log in a user from an API key.

    :status 302: redirect to index on success, to login on failure

    '''
    return bokeh_app.authentication.login_from_apikey()

@bokeh_app.route('/bokeh/register', methods=['GET'])
def register_get():
    ''' Register a new user via a view.

    :status 200: render registration form

    '''
    return bokeh_app.authentication.register_get()

@bokeh_app.route('/bokeh/register', methods=['POST'])
def register_post():
    ''' Register a new user via a submission.

    :status 200: registration result

    '''
    return bokeh_app.authentication.register_post()

@bokeh_app.route('/bokeh/logout')
def logout():
    ''' Log out the current user.

    :status 302: redirect to index

    '''
    return bokeh_app.authentication.logout()

@bokeh_app.route('/bokeh/<docid>/publish', methods=['POST'])
def publish(docid):
    #bokehuser = bokeh_app.current_user()
    doc = docs.Doc.load(bokeh_app.servermodel_storage, docid)
    if not bokeh_app.authentication.can_write_doc(docid):
        return abort(401)
    doc.published = True
    doc.save(bokeh_app.servermodel_storage)
    return jsonify(status='success')



""" Various kinds of markup (static content) widgets.

"""
from __future__ import absolute_import

from ...properties import abstract
from ...properties import Int, String
from ..widget import Widget

@abstract
class Markup(Widget):
    """ Base class for HTML markup widget models. """

class Paragraph(Markup):
    """ A block (paragraph) of text.

    """

    text = String(help="""
    The contents of the widget.
    """)

    width = Int(500, help="""
    The width of the block in pixels.
    """)

    height = Int(400, help="""
    The height of the block in pixels.
    """)


class PreText(Paragraph):
    """ A block (paragraph) of pre-formatted text.

    """



from __future__ import print_function

import numpy as np
import sympy as sy

from bokeh.browserlib import view
from bokeh.document import Document
from bokeh.models.glyphs import Line
from bokeh.models import Plot, DataRange1d, LinearAxis, ColumnDataSource, Grid, Legend
from bokeh.session import Session
from bokeh.models.widgets import Slider, TextInput, HBox, VBox, Dialog

document = Document()
session = Session()
session.use_doc('taylor_server')
session.load_document(document)

xs = sy.Symbol('x')
expr = sy.exp(-xs)*sy.sin(xs)
order = 1

def taylor(fx, xs, order, x_range=(0, 1), n=200):
    x0, x1 = x_range
    x = np.linspace(float(x0), float(x1), n)

    fy = sy.lambdify(xs, fx, modules=['numpy'])(x)
    tx = fx.series(xs, n=order).removeO()

    if tx.is_Number:
        ty = np.zeros_like(x)
        ty.fill(float(tx))
    else:
        ty = sy.lambdify(xs, tx, modules=['numpy'])(x)

    return x, fy, ty

def update_data():
    x, fy, ty = taylor(expr, xs, order, (-2*sy.pi, 2*sy.pi), 200)

    plot.title = "%s vs. taylor(%s, n=%d)" % (expr, expr, order)
    legend.legends = [
        ("%s"         % expr, [line_f_glyph]),
        ("taylor(%s)" % expr, [line_t_glyph]),
    ]
    source.data = dict(x=x, fy=fy, ty=ty)
    slider.value = order

    session.store_document(document)

source = ColumnDataSource(data=dict(x=[], fy=[], ty=[]))

xdr = DataRange1d()
ydr = DataRange1d()

plot = Plot(x_range=xdr, y_range=ydr, plot_width=800, plot_height=400)

line_f = Line(x="x", y="fy", line_color="blue", line_width=2)
line_f_glyph = plot.add_glyph(source, line_f)
plot.add_layout(line_f_glyph)

line_t = Line(x="x", y="ty", line_color="red", line_width=2)
line_t_glyph = plot.add_glyph(source, line_t)
plot.add_layout(line_t_glyph)

xaxis = LinearAxis()
plot.add_layout(xaxis, 'below')

yaxis = LinearAxis()
plot.add_layout(yaxis, 'left')

xgrid = Grid(dimension=0, ticker=xaxis.ticker)
ygrid = Grid(dimension=1, ticker=yaxis.ticker)

legend = Legend(orientation="bottom_left")
plot.add_layout(legend)

def on_slider_value_change(obj, attr, old, new):
    global order
    order = int(new)
    update_data()

def on_text_value_change(obj, attr, old, new):
    try:
        global expr
        expr = sy.sympify(new, dict(x=xs))
    except (sy.SympifyError, TypeError, ValueError) as exception:
        dialog.content = str(exception)
        dialog.visible = True
        session.store_objects(dialog)
    else:
        update_data()

dialog = Dialog(title="Invalid expression")

slider = Slider(start=1, end=20, value=order, step=1, title="Order:")
slider.on_change('value', on_slider_value_change)

text = TextInput(value=str(expr), title="Expression:")
text.on_change('value', on_text_value_change)

inputs = HBox(children=[slider, text])
layout = VBox(children=[inputs, plot, dialog])

document.add(layout)
update_data()

if __name__ == "__main__":
    link = session.object_link(document.context)
    print("Please visit %s to see the plots" % link)
    view(link)
    print("\npress ctrl-C to exit")
    session.poll_document(document)



from __future__ import absolute_import

from os.path import dirname, join

try:
    import pandas as pd
except ImportError as e:
    raise RuntimeError("auto-mpg data requires pandas (http://pandas.pydata.org) to be installed")

autompg = pd.read_csv(join(dirname(__file__), 'auto-mpg.csv'))



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2015, Continuum Analytics, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import logging
log = logging.getLogger(__name__)

from threading import Thread

import zmq

class Subscriber(object):
    def __init__(self, ctx, addrs, wsmanager, timeout=0.1):
        self.ctx = ctx
        self.addrs = addrs
        self.wsmanager = wsmanager
        self.timeout = timeout
        self.kill = False

    def run(self):
        sockets = []
        poller = zmq.Poller()
        for addr in self.addrs:
            log.debug('Subscriber CONNECT: %s' % addr)
            socket = self.ctx.socket(zmq.SUB)
            socket.connect(addr)
            socket.setsockopt_string(zmq.SUBSCRIBE, u"")
            sockets.append(socket)
            poller.register(socket, zmq.POLLIN)
        try:
            while not self.kill:
                socks = dict(poller.poll(self.timeout * 1000))
                for socket, v in socks.items():
                    msg = socket.recv_json()
                    topic, msg, exclude = msg['topic'], msg['msg'], msg['exclude']
                    self.wsmanager.send(topic, msg, exclude=exclude)
        except zmq.ContextTerminated:
            pass
        finally:
            for s in sockets:
                s.close()

    def start(self):
        self.thread = Thread(target=self.run)
        self.thread.start()

    def stop(self):
        self.kill = True



''' Provide decorators help with define Bokeh validation checks.

'''
from __future__ import absolute_import

from functools import partial

from six import string_types

def _validator(code_or_name, validator_type):

    if validator_type == "error":
        from .errors import codes
        from .errors import EXT
    elif validator_type == "warning":
        from .warnings import codes
        from .warnings import EXT
    else:
        pass

    def decorator(func):
        def wrapper(*args, **kw):
            extra = func(*args, **kw)
            if extra is None: return []
            if isinstance(code_or_name, string_types):
                code = EXT
                name = codes[code][0] + ":" + code_or_name
            else:
                code = code_or_name
                name = codes[code][0]
            text = codes[code][1]
            return [(code, name, text, extra)]
        wrapper.validator_type = validator_type
        return wrapper

    return decorator

error = partial(_validator, validator_type="error")
error.__doc__ = ''' Mark a validator method for a Bokeh error condition

Args:
    code_or_name (int or str) : a code from ``bokeh.validation.errors`` or a string label for a custom check

Returns:
    callable : decorator for Bokeh model methods

Examples:

The first example uses a numeric code for a standard error provided in
``bokeh.validation.errors``. This usage is primarily of interest to Bokeh
core developers.

.. code-block:: python

    from bokeh.validation.errors import REQUIRED_RANGES

    @error(REQUIRED_RANGES)
    def _check_no_glyph_renderers(self):

The second example shows how a custom warning check can be implemented by
passing an arbitrary string label to the decorator. This usage is primarily
of interest to anyone extending Bokeh with their own custom models.

.. code-block:: python

    @error("MY_CUSTOM_WARNING")
    def _check_my_custom_warning(self):


'''

warning = partial(_validator, validator_type="warning")
warning.__doc__ = ''' Mark a validator method for a Bokeh error condition

Args:
    code_or_name (int or str) : a code from ``bokeh.validation.errors`` or a string label for a custom check

Returns:
    callable : decorator for Bokeh model methods

Examples:

The first example uses a numeric code for a standard warning provided in
``bokeh.validation.warnings``. This usage is primarily of interest to Bokeh
core developers.

.. code-block:: python

    from bokeh.validation.warnings import NO_GLYPH_RENDERERS

    @warning(NO_GLYPH_RENDERERS)
    def _check_no_glyph_renderers(self):

The second example shows how a custom warning check can be implemented by
passing an arbitrary string label to the decorator. This usage is primarily
of interest to anyone extending Bokeh with their own custom models.

.. code-block:: python

    @warning("MY_CUSTOM_WARNING")
    def _check_my_custom_warning(self):



'''



""" This is the Bokeh charts testing interface.

"""
#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2014, Continuum Analytics, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
from __future__ import absolute_import

from collections import OrderedDict
import unittest

import numpy as np
from numpy.testing import assert_array_equal

from bokeh._legacy_charts import Donut

from ._utils import create_chart

#-----------------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------------

class TestDonut(unittest.TestCase):

    def test_supported_input(self):
        xyvalues = OrderedDict()
        # TODO: Fix bug for donut breaking when inputs that are not float
        xyvalues['python'] = [2., 5., 3.]
        xyvalues['pypy'] = [4., 1., 4.]
        xyvalues['jython'] = [6., 4., 3.]

        xyvalues_int = OrderedDict()
        for k, values in xyvalues.items():
            xyvalues_int[k] = [int(val) for val in values]

        for xyvalues in [xyvalues, xyvalues_int]:
            cat = ["sets", "dicts", "odicts"]
            start = [0, 2.3561944901923448, 4.3196898986859651]
            end = [2.3561944901923448, 4.3196898986859651, 6.2831853071795862]
            colors = ['#f22c40', '#5ab738', '#407ee7']

            # TODO: Chart is not working with DataFrames anymore.
            #       Fix it and add test case for , pd.DataFrame(xyvalues)
            for i, _xy in enumerate([xyvalues]):
                _chart = create_chart(Donut, _xy, cat=cat)
                builder = _chart._builders[0]
                self.assertEqual(builder._groups, cat)
                assert_array_equal(builder._data['start'], start)
                assert_array_equal(builder._data['end'], end)
                assert_array_equal(builder._data['colors'], colors)

                # TODO: Test for external ring source values is missing as it needs
                #       some refactoring to expose those values calculation

        lvalues = [[2., 5., 3.], [4., 1., 4.], [6., 4., 3.]]
        lvalues_int = [[2, 5, 3], [4, 1, 4], [6, 4, 3]]
        for lvalues in [lvalues, lvalues_int]:
            for i, _xy in enumerate([lvalues, np.array(lvalues)]):
                _chart = create_chart(Donut, _xy, cat=cat)
                builder = _chart._builders[0]
                self.assertEqual(builder._groups, cat)
                assert_array_equal(builder._data['start'], start)
                assert_array_equal(builder._data['end'], end)
                assert_array_equal(builder._data['colors'], colors)

                # TODO: Test for external ring source values is missing as it needs
                #       some refactoring to expose those values calculation



from collections import OrderedDict

from bokeh.sampledata import us_counties, unemployment
from bokeh.plotting import figure, show, output_file, ColumnDataSource
from bokeh.models import HoverTool

state = "tx"

county_xs=[
    us_counties.data[code]['lons'] for code in us_counties.data
    if us_counties.data[code]['state'] == state
]
county_ys=[
    us_counties.data[code]['lats'] for code in us_counties.data
    if us_counties.data[code]['state'] == state
]

colors = ["#F1EEF6", "#D4B9DA", "#C994C7", "#DF65B0", "#DD1C77", "#980043"]

county_colors = []
county_names = []
county_rates = []
for county_id in us_counties.data:
    if us_counties.data[county_id]['state'] != state:
        continue
    rate = unemployment.data[county_id]
    idx = min(int(rate/2), 5)
    county_colors.append(colors[idx])
    county_names.append(us_counties.data[county_id]['name'])
    county_rates.append(rate)

source = ColumnDataSource(
    data = dict(
        x=county_xs,
        y=county_ys,
        color=county_colors,
        name=county_names,
        rate=county_rates,
    )
)


output_file("texas.html", title="texas.py example")

TOOLS="pan,wheel_zoom,box_zoom,reset,hover,save"

p = figure(title="Texas Unemployment 2009", tools=TOOLS)

p.patches('x', 'y',
    fill_color='color', fill_alpha=0.7,
    line_color="white", line_width=0.5,
    source=source)

hover = p.select(dict(type=HoverTool))
hover.point_policy = "follow_mouse"
hover.tooltips = OrderedDict([
    ("Name", "@name"),
    ("Unemployment rate)", "@rate%"),
    ("(Long, Lat)", "($x, $y)"),
])

show(p)



""" Thoroughly document Bokeh property attributes.

The ``bokeh-prop`` directive generates useful type information
for the property attribute, including cross links to the relevant
property types. Additionally, any per-attribute docstrings are
also displayed.

Usage
-----

This directive takes the path to an attribute on a Bokeh
model class as an argument::

    .. bokeh-prop:: bokeh.sphinxext.sample.Bar.thing

Examples
--------

For the following definition of ``bokeh.sphinxext.sample.Bar``::

    class Bar(PlotObject):
        ''' This is a Bar model. '''
        thing = List(Int, help="doc for thing")

the above usage yields the output:

----

    .. bokeh-prop:: bokeh.sphinxext.sample.Bar.thing

"""
from __future__ import absolute_import, print_function

import importlib

from docutils import nodes
from docutils.statemachine import ViewList

import textwrap

import jinja2

from sphinx.util.compat import Directive
from sphinx.util.nodes import nested_parse_with_titles

from bokeh.plot_object import Viewable
import bokeh.properties


PROP_TEMPLATE = jinja2.Template(u"""
.. attribute:: {{ name }}
    :module: {{ module }}

    *property type:* {{ type_info }}

    {% if doc %}{{ doc|indent(4) }}{% endif %}

""")

PROP_NAMES = [
    name for name, cls in bokeh.properties.__dict__.items()
    if isinstance(cls, type) and issubclass(cls, bokeh.properties.Property)
]
PROP_NAMES.sort(reverse=True, key=len)

class BokehPropDirective(Directive):

    has_content = True
    required_arguments = 1

    def run(self):

        prop_path = self.arguments[0]
        module_path, model_name, prop_name = prop_path.rsplit('.', 2)

        try:
            module = importlib.import_module(module_path)
        except ImportError:
            pass

        model = getattr(module, model_name, None)
        if model is None:
            pass

        if type(model) != Viewable:
            pass

        model_obj = model()

        prop = getattr(model_obj.__class__, prop_name)

        type_info = self._get_type_info(prop)

        rst_text = PROP_TEMPLATE.render(
            name=prop_name,
            module=module_path,
            type_info=type_info,
            doc="" if prop.__doc__ is None else textwrap.dedent(prop.__doc__),
        )

        result = ViewList()
        for line in rst_text.split("\n"):
            result.append(line, "<bokeh-prop>")
        node = nodes.paragraph()
        node.document = self.state.document
        nested_parse_with_titles(self.state, result, node)
        return node.children

    def _get_type_info(self, prop):
        desc = str(prop)
        template = ":class:`~bokeh.properties.%s`\ "
        # some of the property names are substrings of other property names
        # so first go through greedily replacing the longest possible match
        # with a unique id (PROP_NAMES is reverse sorted by length)
        for i, name in enumerate(PROP_NAMES):
            desc = desc.replace(name, "__ID%d" % i)
        # now replace the unique id with the corresponding prop name. Go in
        # reverse to make sure replacements are greedy
        for i in range(len(PROP_NAMES)-1, 0, -1):
            name = PROP_NAMES[i]
            desc = desc.replace("__ID%d" % i, template % name)
        return desc

def setup(app):
    app.add_directive_to_domain('py', 'bokeh-prop', BokehPropDirective)



from bokeh.plotting import figure, show, output_file, vplot

N = 4000

factors = ["a", "b", "c", "d", "e", "f", "g", "h"]
x0 = [0, 0, 0, 0, 0, 0, 0, 0]
x =  [50, 40, 65, 10, 25, 37, 80, 60]

output_file("categorical.html", title="categorical.py example")

p1 = figure(title="Dot Plot", tools="resize,save", y_range=factors, x_range=[0,100])

p1.segment(x0, factors, x, factors, line_width=2, line_color="green", )
p1.circle(x, factors, size=15, fill_color="orange", line_color="green", line_width=3, )

factors = ["foo", "bar", "baz"]
x = ["foo", "foo", "foo", "bar", "bar", "bar", "baz", "baz", "baz"]
y = ["foo", "bar", "baz", "foo", "bar", "baz", "foo", "bar", "baz"]
colors = [
    "#0B486B", "#79BD9A", "#CFF09E",
    "#79BD9A", "#0B486B", "#79BD9A",
    "#CFF09E", "#79BD9A", "#0B486B"
]

p2 = figure(title="Categorical Heatmap", tools="resize,hover,save",
    x_range=factors, y_range=factors)

p2.rect(x, y, color=colors, width=1, height=1)

show(vplot(p1, p2))  # open a browser



from __future__ import print_function

from numpy import pi, arange, sin, cos

from bokeh.browserlib import view
from bokeh.document import Document
from bokeh.models.glyphs import Circle
from bokeh.models import (
    Plot, DataRange1d, LinearAxis, Grid,
    ColumnDataSource, PanTool, WheelZoomTool
)
from bokeh import session

document = Document()
session = session.Session()
session.use_doc('glyph2_server')
session.load_document(document)

x = arange(-2*pi, 2*pi, 0.1)
y = sin(x)
r = (cos(x)+1) * 6 + 6

source = ColumnDataSource(data=dict(x=x, y=y, r=r))

xdr = DataRange1d()
ydr = DataRange1d()

plot = Plot(x_range=xdr, y_range=ydr, min_border=80)

circle = Circle(
    x="x", y="y", size="r",
    fill_color="red", line_color="black"
)
plot.add_glyph(source, circle)

xaxis = LinearAxis()
plot.add_layout(xaxis, 'below')

yaxis = LinearAxis()
plot.add_layout(yaxis, 'left')

plot.add_layout(Grid(dimension=0, ticker=xaxis.ticker))
plot.add_layout(Grid(dimension=1, ticker=yaxis.ticker))

plot.add_tools(PanTool(), WheelZoomTool())

document.add(plot)
session.store_document(document)

link = session.object_link(document.context)
print ("please visit %s to see plots" % link)
view(link)




from bokeh.sampledata.glucose import data
from bokeh.plotting import figure, show, output_file
from bokeh.models import BoxAnnotation

output_file("box_annotation.html", title="box_annotation.py example")

TOOLS = "pan,wheel_zoom,box_zoom,reset,save"

#reduce data size
data = data.ix['2010-10-06':'2010-10-13']

p = figure(x_axis_type="datetime", tools=TOOLS)

p.line(data.index.to_series(), data['glucose'],
       line_color="gray", line_width=1, legend="glucose")

low_box = BoxAnnotation(plot=p, top=80, fill_alpha=0.1, fill_color='red')
mid_box = BoxAnnotation(plot=p, bottom=80, top=180, fill_alpha=0.1, fill_color='green')
high_box = BoxAnnotation(plot=p, bottom=180, fill_alpha=0.1, fill_color='red')

p.renderers.extend([low_box, mid_box, high_box])

p.title = "Glucose Range"
p.xgrid[0].grid_line_color=None
p.ygrid[0].grid_line_alpha=0.5
p.xaxis.axis_label = 'Time'
p.yaxis.axis_label = 'Value'

show(p)



''' The query module provides functions for searching Bokeh object
graphs for objects that match specified criteria.

Queries are specified as selectors similar to MongoDB style query
selectors.

Examples::

    # find all objects with type "grid"
    find(p, {'type': 'grid'})

    # find all objects with type "grid" or "axis"
    find(p, {OR: [
        {'type': 'grid'}, {'type': 'axis'}
    ]})

    # same query, using IN operator
    find(p, {'type': {IN: ['grid', 'axis']})

    # find all plot objects on the 'left' layout of the Plot
    list(find(p, {'layout': 'left'}, {'plot': p}))

    # find all subplots in column 0
    find(p, {type: 'plot', 'column: 0}, {'gridplot': p})

    # find all subplots the last row
    find(p, {type: 'plot', 'row': -1}, {'gridplot': p})

'''

from __future__ import absolute_import

from six import string_types

class OR(object): pass

class IN(object): pass
class GT(object): pass
class LT(object): pass
class EQ(object): pass
class GEQ(object): pass
class LEQ(object): pass
class NEQ(object): pass


def match(obj, selector, context=None):
    ''' Test whether a particular object matches a given
    selector.

    Args:
        obj (PlotObject) : object to Test
        selector (JSON-like) : query selector
            See module docs for details

    Returns:
        bool : True if the object matches, False otherwise

    There are two selector keys that are handled specially. The first
    is 'type', which will do an isinstance check::

        >>> from bokeh.plotting import line
        >>> from bokeh.models import Axis
        >>> p = line([1,2,3], [4,5,6])
        >>> len(list(p.select({'type': Axis})))
        2

    There is also a 'tags' attribute that `PlotObject` objects have,
    that is a list of user-supplied values. The 'tags' selector key can
    be used to query against this list of tags. An object matches if
    any of the tags in the selector match any of the tags on the
    object::

        >>> from bokeh.plotting import line
        >>> from bokeh.models import Axis
        >>> p = line([1,2,3], [4,5,6])
        >>> p.tags = ["my plot", 10]
        >>> len(list(p.select({'tags': "my plot"})))
        1
        >>> len(list(p.select({'tags': ["my plot", 10]})))
        1

    '''
    context = context or {}
    for key, val in selector.items():

        # test attributes
        if isinstance(key, string_types):

            # special case 'type'
            if key == "type":
                # type supports IN, check for that first
                if isinstance(val, dict) and list(val.keys()) == [IN]:
                    if not any(isinstance(obj, x) for x in val[IN]): return False
                # otherwise just check the type of the object against val
                elif not isinstance(obj, val): return False

            # special case 'tag'
            elif key == 'tags':
                if isinstance(val, string_types):
                    if val not in obj.tags: return False
                else:
                    try:
                        if not set(val) & set(obj.tags): return False
                    except TypeError:
                        if val not in obj.tags: return False

            # if the object doesn't have the attr, it doesn't match
            elif not hasattr(obj, key): return False

            # if the value to check is a dict, recurse
            else:
                attr = getattr(obj, key)
                if callable(attr):
                    try:
                        if not attr(val, **context): return False
                    except:
                        return False

                elif isinstance(val, dict):
                    if not match(attr, val, context): return False

                else:
                    if attr != val: return False

        # test OR conditionals
        elif key is OR:
            if not _or(obj, val): return False

        # test operands
        elif key in _operators:
            if not _operators[key](obj, val): return False

        else:
            raise ValueError("malformed query selector")

    return True


def find(objs, selector, context=None):
    ''' Query an object and all of its contained references
    and yield objects that match the given selector.

    Args:
        obj (PlotObject) : object to query
        selector (JSON-like) : query selector
            See module docs for details

    Yields:
        PlotObject : objects that match the query

    Examples:

    '''
    return (obj for obj in objs if match(obj, selector, context))


def _or(obj, selectors):
    return any(match(obj, selector) for selector in selectors)


_operators = {
   IN:  lambda x, y: x in y,
   GT:  lambda x, y: x > y,
   LT:  lambda x, y: x < y,
   EQ:  lambda x, y: x == y,
   GEQ: lambda x, y: x >= y,
   LEQ: lambda x, y: x <= y,
   NEQ: lambda x, y: x != y,
}



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2015, Continuum Analytics, Inc. All rights reserved.
#
# Powered by the Bokeh Development Team.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
""" Utilities for writing plugins.

This is different from bokeh.pluginutils because these are ways of
patching routes and objects directly into the bokeh server. You
would run this type of code using the --script option

"""

from __future__ import absolute_import

import uuid

from bokeh.exceptions import DataIntegrityException
from bokeh.resources import Resources
from flask import abort, render_template

from ..app import bokeh_app
from ..views.backbone import init_bokeh
from ..views.main import _makedoc

def object_page(prefix):
    """ Decorator for a function which turns an object into a web page

    from bokeh.server.app import bokeh_app
    @bokeh_app.route("/myapp")
    @object_page("mypage")
    def make_object():
        #make some bokeh object here
        return obj

    This decorator will
      - create a randomized title for a bokeh document using the prefix
      - initialize bokeh plotting libraries to use that document
      - call the function you pass in, add that object to the plot context
      - render that object in a web page

    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            ## setup the randomly titled document
            docname = prefix + str(uuid.uuid4())
            bokehuser = bokeh_app.current_user()
            try:
                doc = _makedoc(bokeh_app.servermodel_storage, bokehuser, docname)
                doc.published = True
                doc.save(bokeh_app.servermodel_storage)
            except DataIntegrityException as e:
                return abort(409, e.message)
            docid = doc.docid
            clientdoc = bokeh_app.backbone_storage.get_document(docid)

            ## initialize our plotting APIs to use that document

            init_bokeh(clientdoc)
            obj = func(*args, **kwargs)
            clientdoc.add(obj)
            bokeh_app.backbone_storage.store_document(clientdoc)
            if hasattr(obj, 'extra_generated_classes'):
                extra_generated_classes = obj.extra_generated_classes
            else:
                extra_generated_classes = []

            resources = Resources()
            return render_template("oneobj.html",
                                   elementid=str(uuid.uuid4()),
                                   docid=docid,
                                   objid=obj._id,
                                   hide_navbar=True,
                                   extra_generated_classes=extra_generated_classes,
                                   public='true',
                                   loglevel=resources.log_level)
        wrapper.__name__ = func.__name__
        return wrapper

    return decorator



from bokeh.plotting import figure
from bokeh.models import Range1d
from bokeh.embed import components
from jinja2 import Template
import webbrowser
import os
import six

# create some data
x1 = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
y1 = [0, 8, 2, 4, 6, 9, 5, 6, 25, 28, 4, 7]
x2 = [2, 5, 7, 15, 18, 19, 25, 28, 9, 10, 4]
y2 = [2, 4, 6, 9, 15, 18, 0, 8, 2, 25, 28]
x3 = [0, 1, 0, 8, 2, 4, 6, 9, 7, 8, 9]
y3 = [0, 8, 4, 6, 9, 15, 18, 19, 19, 25, 28]

# select the tools we want
TOOLS="pan,wheel_zoom,box_zoom,reset,save"

# the red and blue graphs will share this data range
xr1 = Range1d(start=0, end=30)
yr1 = Range1d(start=0, end=30)

# only the green will use this data range
xr2 = Range1d(start=0, end=30)
yr2 = Range1d(start=0, end=30)

# build our figures
p1 = figure(x_range=xr1, y_range=yr1, tools=TOOLS, plot_width=300, plot_height=300)
p1.scatter(x1, y1, size=12, color="red", alpha=0.5)

p2 = figure(x_range=xr1, y_range=yr1, tools=TOOLS, plot_width=300, plot_height=300)
p2.scatter(x2, y2, size=12, color="blue", alpha=0.5)

p3 = figure(x_range=xr2, y_range=yr2, tools=TOOLS, plot_width=300, plot_height=300)
p3.scatter(x3, y3, size=12, color="green", alpha=0.5)

# plots can be a single PlotObject, a list/tuple, or even a dictionary
plots = {'Red': p1, 'Blue': p2, 'Green': p3}

script, div = components(plots)

template = Template('''<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset="utf-8">
        <title>Bokeh Scatter Plots</title>
        <style> div{float: left;} </style>
        <link rel="stylesheet" href="http://cdn.pydata.org/bokeh/release/bokeh-0.9.0.min.css" type="text/css" />
        <script type="text/javascript" src="http://cdn.pydata.org/bokeh/release/bokeh-0.9.0.min.js"></script>
        {{ script }}
    </head>
    <body>
    {% for key in div.keys() %}
        {{ div[key] }}
    {% endfor %}
    </body>
</html>
''')

html_file = 'embed_multiple.html'
with open(html_file, 'w') as textfile:
    textfile.write(template.render(script=script, div=div))
url = 'file:{}'.format(six.moves.urllib.request.pathname2url(os.path.abspath(html_file)))
webbrowser.open(url)



from matplotlib.collections import PolyCollection
import matplotlib.pyplot as plt
import numpy as np

from bokeh import mpl
from bokeh.plotting import output_file, show

# Generate data. In this case, we'll make a bunch of center-points and generate
# verticies by subtracting random offsets from those center-points
numpoly, numverts = 100, 4
centers = 100 * (np.random.random((numpoly, 2)) - 0.5)
offsets = 10 * (np.random.random((numverts, numpoly, 2)) - 0.5)
verts = centers + offsets
verts = np.swapaxes(verts, 0, 1)

# In your case, "verts" might be something like:
# verts = zip(zip(lon1, lat1), zip(lon2, lat2), ...)
# If "data" in your case is a numpy array, there are cleaner ways to reorder
# things to suit.

facecolors = ['red', 'green', 'blue', 'cyan', 'yellow', 'magenta', 'black']
edgecolors = ['cyan', 'yellow', 'magenta', 'black', 'red', 'green', 'blue']
widths = [5, 10, 20, 10, 5]


# Make the collection and add it to the plot.
col = PolyCollection(verts, facecolor=facecolors, edgecolor=edgecolors,
                     linewidth=widths, linestyle='--', alpha=0.5)

ax = plt.axes()
ax.add_collection(col)

plt.xlim([-60, 60])
plt.ylim([-60, 60])

plt.title("MPL-PolyCollection support in Bokeh")

output_file("polycollection.html")

show(mpl.to_bokeh())



import gzip,time,sys, os, random
here = os.path.dirname(os.path.realpath(__file__))
filepath = os.path.join(here, 'sim_tweets.csv')

header = 'text,lat,lon\n'
with file(filepath, 'w') as f:
    f.write(header)

tweets = []
lons = []
lats = []
# get some saved simulated tweet as base
for i, line in enumerate(gzip.open(os.path.join(here, "simplified_geo_tweets.gz"))):
    if i > 0:
        line = line.strip()
        tweet = line.split(',')
        if len(tweet) == 3:

            tweets.append(tweet)
            lons.append(float(tweet[1]))
            lats.append(float(tweet[2]))
            with file(filepath, 'a') as f:
                f.write("%s\n" % line)
            time.sleep(0.1)

delta = 1
lons = [min(lons)-delta, max(lons)+delta]
lats = [min(lats)-delta, max(lats)+delta]
# start infinite retweets from random locations insite the original tweets area
while True:
    for tweet in tweets:
        tweet[1] = random.uniform(*lons)
        tweet[2] = random.uniform(*lats)
        line = ','.join(str(x) for x in tweet)

        with file(filepath, 'a') as f:
            f.write("%s\n" % line)
        time.sleep(0.1)


from __future__ import absolute_import, print_function
from ..plotting import figure, curdoc
from ..plot_object import PlotObject
from ..models import ServerDataSource,  GlyphRenderer, Range1d, Color
from ..properties import (Instance, Any, Either,
                          Int, Float, List, Bool, String)
from six import get_function_code, string_types

import bokeh.colors as colors
import numpy as np
import math

import logging
logger = logging.getLogger(__file__)

_AR_VERSION = "0.5.1"
_AR_MESSAGE = """
---------------------------------------------------
Error loading the abstract_rendering package.

To use the ar_downsample module, you must install the
abstract rendering framework.
This can be installed with conda, pip or by
cloning from https://github.com/ContinuumIO/abstract_rendering
Questions and feedback can be directed to
Joseph Cottam (jcottam@indiana.edu)
-------------------------------------------------------
"""

# Tell pyflakes that these names exists
ari = ar = categories = contour = infos = general = glyphset = npg = numeric = util = None

def _loadAR():
    """
    Utility to load abstract rendering (AR) and checks for version match.
    Keeps the import from occurring
    unless you actually try to use AR.  MUST be called before actually
    calling any AR package items (typically only invoked on the server
    in response to an AR-reliant plot request).

    This is more complex than just an import because
    AR is exposed as several backbone modules.  If the AR modules
    were directly imported, then errors would occur whenever ar_downsample
    is imported.  This causes error messages on python client side (where
    AR isn't actually needed) and on the server side even when AR isn't used.
    Since the AR modules are used throughout this module, just importing at
    use point inside this module is cumbersome.  Using 'globals()' and
    importlib allows this method to be called before any AR proper items
    are used but still have the imports appear at the module level.
    """
    try:
        from importlib import import_module
        globals()["ari"] = import_module("abstract_rendering")
    except:
        print(_AR_MESSAGE)
        raise

    expected = dict(zip(["major", "minor", "micro"], map(int, _AR_VERSION.split("."))))
    if not hasattr(ari, "__version_info__"):
        raise ImportError("Abstract rendering version not found; import aborted.")

    if (ari.__version_info__["major"] != expected["major"]
        or ari.__version_info__["minor"] != expected["minor"]
        or ari.__version_info__["micro"] < expected["micro"]):
            raise ImportError("Abstract rendering version mismatched." +
                              "Expecting at least {0}, found {1}".format(_AR_VERSION, ari.__version__))

    try:
        globals()["ar"] = import_module("abstract_rendering.core")
        globals()["categories"] = import_module("abstract_rendering.categories")
        globals()["contour"] = import_module("abstract_rendering.contour")
        globals()["infos"] = import_module("abstract_rendering.infos")
        globals()["general"] = import_module("abstract_rendering.general")
        globals()["glyphset"] = import_module("abstract_rendering.glyphset")
        globals()["npg"] = import_module("abstract_rendering.numpyglyphs")
        globals()["numeric"] = import_module("abstract_rendering.numeric")
        globals()["util"] = import_module("abstract_rendering.util")
    except:
        print(_AR_MESSAGE)
        raise


class Proxy(PlotObject):
    """
    Proxy objects stand in for the abstract rendering (AR) configuration
    classes. Basically, the AR implementation doesn't rely on Bokeh, so
    it doesn't know about the properties BUT the Bokeh needs be able to
    construct/modify/inspect AR configurations.  Proxy classes hold the
    relevant parameters for constructing AR classes in a way that Bokeh
    can inspect. Furthermore, 'reify' produces an AR class from a
    proxy instance.
    """
    def reify(self, **kwargs):
        raise NotImplementedError("Unimplemented")


# ------ Aggregators -------------
class Sum(Proxy):
    "Add up all incoming values"
    def reify(self, **kwargs):
        return numeric.Sum()


class Count(Proxy):
    "Count how many items there are (ignores values)"
    def reify(self, **kwargs):
        glyphs = kwargs.get('glyphset', None)
        if isinstance(glyphs, npg.Glyphset):
            return npg.PointCount()
        else:
            return numeric.Count()


class CountCategories(Proxy):
    "Count how many items there are in provided categories"
    def reify(self, **kwargs):
        glyphs = kwargs.get('glyphset', None)
        if isinstance(glyphs, npg.Glyphset):
            return npg.PointCountCategories()
        else:
            return categories.CountCategories()


# ------ Infos ---------
class Const(Proxy):
    "Return a single value"
    val = Any()

    def reify(self, **kwargs):
        return infos.const(self.val)


class Encode(Proxy):
    "Convert a set of values to numeric codes."
    cats = List(Any)
    defcat = Int(-1)

    def reify(self, **kwargs):
        return infos.encode(self.cats, defcat=self.defcat)


class AutoEncode(Proxy):
    "Convert a set of values to numeric codes."

    def reify(self, **kwargs):
        return infos.AutoEncode()


# ----- Shaders ---------
# Out types to support:
#   image -- grid of values
#   rgb_image -- grid of colors
#   multi_line -- multi-segment lines

class Shader(Proxy):
    def __add__(self, other):
        "Build a sequence of shaders"
        return Seq(first=self, second=other)

    def reformat(self, result, *args):
        """
        Modify the standard AR shader output to one that Bokeh can use.
        Response to 'None' should produce an empty basic dataset.

        The default behavior for None is [[0]] (most shaders produce images).
        The default behavior for all other values is identity.
        """
        return np.array([[0]]) if result is None else result


class Seq(Shader):
    "Sequence of shaders"
    first = Instance(Shader)
    second = Instance(Shader)

    def reify(self, **kwargs):
        return self.first.reify(**kwargs) + self.second.reify(**kwargs)

    @property
    def out(self):
        return self.second.out

    def reformat(self, result, *args):
        return self.second.reformat(result, *args)


class Id(Shader):
    "Identity shader.  A safe default."
    out = "image"

    def reify(self, **kwargs):
        return general.Id()


class BinarySegment(Shader):
    """
    Divide the input space into two regions.

    divider - Value that divides the two regions
    high - Value for regions equal to or above the divider
    low - Value for cells below divider
    """
    out = "image"
    high = Any
    low = Any
    divider = Any   # TODO: Restrict to numbers...

    def reify(self, **kwargs):
        return numeric.BinarySegment(self.low, self.high, self.divider)


class Interpolate(Shader):
    "Interpolate between high and low number"
    out = "image"
    high = Any     # TODO: Restrict to numbers...
    low = Any

    def reify(self, **kwargs):
        return numeric.Interpolate(self.low, self.high)


class ImageShader(Shader):
    out = "image_rgb"

    def _reformatColor(self, color):
        if isinstance(color, tuple) or isinstance(color, list):
            parts = len(color)
            if parts == 3:
                return tuple(color)+(255,)
            if parts == 4:
                return tuple(color[0:3]) + (min(abs(color[3])*255, 255),)
            raise ValueError("Improperly formatted tuple for color %s" % color)

        if isinstance(color, string_types):
            if color[0] == "#":
                color = color.lstrip('#')
                lv = len(color)
                rgb = tuple(int(color[i:i + lv // 3], 16) for i in range(0, lv, lv // 3))
                return [rgb[0], rgb[1], rgb[2], 255]
            else:
                try:
                    rgb = getattr(colors, color).to_rgb()
                    return [rgb.r, rgb.g, rgb.b, 255]
                except:
                    raise ValueError("Unknown color string %s" % color)

    def reformat(self, image, *args):
        if image is None:
            return np.array([[0]])
        else:
            if not image.flags.contiguous:
                # TODO: Numpyglyphs *sometimes* return a non-continugous array?
                image = np.ascontiguousarray(image)
            return image.view(dtype=np.int32).reshape(image.shape[0:2])


class ToCounts(Shader):
    "Convert count-of-categories to just counts"
    out = "image"

    def reify(self, **kwargs):
        return categories.ToCounts()


class NonZeros(Shader):
    "How many non-zero categories are there?"
    out = "image"

    def reify(self, **kwargs):
        return categories.NonZeros()


class Ratio(Shader):
    "Ratio of some category to total of all categories."
    out = "image"
    focus = Int(-1)

    def reify(self, **kwargs):
        return categories.Ratio(focus=self.focus)


class HDAlpha(ImageShader):
    colors = List(Color)
    background = Color((255, 255, 255, 0))
    alphamin = Float(.3)
    log = Bool(False)
    base = Int(10)

    def reify(self, **kwargs):
        colors = map(self._reformatColor, self.colors)
        return categories.HDAlpha(
                colors,
                background=self._reformatColor(self.background),
                alphamin=self.alphamin,
                log=self.log,
                logbase=self.base)


class InterpolateColor(ImageShader):
    """
    Interpolate between a high and low color

    * high - High color (default: red)
    * low - Low color (default: white)
    * reserve - Color for empty values (default: transparent)
    * empty - Empty value (default: 0)

    """
    # TODO: Make color format conversion fluid
    #       ...possibly by some change to properties.Color
    out = "image_rgb"
    high = Color((255, 0, 0, 1))
    low = Color((255, 255, 255, 1))
    reserve = Color((255, 255, 255, 0))
    empty = Any(0)

    def reify(self, **kwargs):
        return numeric.InterpolateColors(
            self._reformatColor(self.low),
            self._reformatColor(self.high),
            reserve=self._reformatColor(self.reserve),
            empty=self.empty)


class Sqrt(Shader):
    "Square root of all values"
    out = "image"

    def reify(self, **kwargs):
        return numeric.Sqrt()


class Cuberoot(Shader):
    "Cube root of all values"
    out = "image"

    def reify(self, **kwargs):
        return numeric.Cuberoot()


class Log(Shader):
    "Log (base 10) of values"
    out = "image"

    def reify(self, **kwargs):
        return npg.Log10()


class Spread(Shader):
    """Spread values out in a regular pattern from their origin."""
    # TODO: Add shape parameter

    out = "image"
    factor = Int
    shape = String("circle")
    anti_alias = Bool(False)

    def reify(self, **kwargs):
        return npg.Spread(factor=self.factor,
                          shape=self.shape,
                          anti_alias=self.anti_alias)


class Contour(Shader):
    """
    ISO Contours

    levels -- Either how many ISO contours to make (int)
              or the exact contour levels (list).
              Both cases indicate how many contour lines to create.
    """
    out = "multi_line"
    levels = Either(Int, List(Float), default=5)
    palette = List(Color)

    def reify(self, **kwargs):
        return contour.Contour(levels=self.levels, points=False)

    def reformat(self, contours, *args):
        if contours is None:
            return {'xxs': [], 'yys': [], 'levels': []}

        xxs = []
        yys = []
        levels = sorted(contours.keys())
        colors = []
        (xmin, ymin) = args

        # Re-arrange results and project xs/ys back to the data space
        for (level, color) in zip(levels, self.palette):
            for trace in contours[level]:
                (xs, ys) = trace
                xs = xs+xmin
                ys = ys+ymin
                xxs.append(xs)
                yys.append(ys)
                colors.append(color)

        return {'levels': levels,
                'colors': colors,
                'xs': xxs,
                'ys': yys}


# ------------------  Control Functions ----------------
def replot(plot,
           agg=Count(), info=Const(val=1), shader=Id(),
           remove_original=True,
           plot_opts=None, **kwargs):
    """
    Treat the passed plot as an base plot for abstract rendering, generate the
    proper Bokeh plot based on the passed parameters.

    This is a convenience for:
    > src=source(plot, ...)
    > <plot-type>(source=src, <params>, **ar.mapping(src))

    Transfers plot title, width, height from the original plot

    plot -- Plot to based new plot off of
    remove_original -- Remove the source plot from the document (default: true)
    plot_opts -- Options passed directly to soure.
                   This parameter only needs to be used if there is a name
                   conflict bewteen  target plot type and
                   source options (see kwargs note)
    **kwargs -- Arugments for the plot or source as needed.
                If in conflict, a kwarg will be applied to the source function.
                If this causes incorrecte behavior, the plot arguments may
                be put in the plot_opts parameter instead.
    returns -- A new plot
    """

    # Remove the base plot (if requested)
    if remove_original and plot in curdoc().context.children:
        curdoc().context.children.remove(plot)

    # Sift kwargs
    source_opts = {}
    for key in ServerDataSource().vm_props():
        if key in kwargs:
            source_opts[key] = kwargs.pop(key)

    for name in get_function_code(source).co_varnames:
        if name in kwargs:
            source_opts[name] = kwargs.pop(name)

    # Transfer specific items from the source plot, then updates from kwargs
    fig_opts = {}
    fig_opts['plot_width'] = kwargs.pop('plot_width', plot.plot_width)
    fig_opts['plot_height'] = kwargs.pop('plot_height', plot.plot_height)
    fig_opts['title'] = kwargs.pop('title', plot.title)

    src = source(plot, agg, info, shader, **source_opts)
    plot_opts = plot_opts or {}
    plot_opts.update(mapping(src))

    new_plot = figure(**fig_opts)
    if shader.out == "image":
        new_plot.image(source=src, **plot_opts)
    elif shader.out == "image_rgb":
        new_plot.image_rgba(source=src, **plot_opts)
    elif shader.out == "multi_line":
        new_plot.multi_line(source=src, **plot_opts)
    else:
        raise ValueError("Unhandled output type %s" % shader.out)

    return new_plot


def _renderer(plot):
    """
    Acquire the renderer of interest.
    More specifically, the first child of the plot
    that is both a renderer and related to a server data source.

    TODO: How to be more specific about what to do AR on?
          Currently just takes the first renderer with a server data source
    """
    return [r for r in plot.renderers
            if (isinstance(r, GlyphRenderer)
                and isinstance(r.data_source, ServerDataSource))][0]

# TODO: just use glyph instead of obsolete glyphspec
def get_glyphspec(glyph):
    spec = glyph.vm_serialize()
    spec['type'] = glyph.__view_model__
    return spec


def source(plot,
           agg=Count(), info=Const(val=1), shader=Id(),
           points=False, balancedZoom=False, **kwargs):
    # Acquire information from renderer...
    rend = _renderer(plot)
    datasource = rend.data_source
    kwargs['data_url'] = datasource.data_url
    kwargs['expr'] = datasource.expr
    kwargs['namespace'] = datasource.namespace
    # TODO: Use reformat here?
    if shader.out == "image" or shader.out == "image_rgb":
        kwargs['data'] = {'image': [],
                          'x': [0],
                          'y': [0],
                          'global_x_range': [0, 50],
                          'global_y_range': [0, 50],
                          'global_offset_x': [0],
                          'global_offset_y': [0],
                          'dw': [1],
                          'dh': [1],
                          'render_state': {}}
    elif shader.out == "multi_line":
        kwargs['data'] = {'xs': [[]],
                          'ys': [[]],
                          'colors': [],
                          'render_state': {}}
    else:
        raise ValueError("Unrecognized shader output type %s" % shader.out)

    kwargs['transform'] = {
        'resample': "abstract rendering",
        'auto_bounds' : True,
        'agg': agg,
        'info': info,
        'shader': shader,
        'glyphspec': rend.glyph,
        'balancedZoom': balancedZoom,
        'points': points}

    return ServerDataSource(**kwargs)


def mapping(source):
    "Setup property mapping dictionary from source to output glyph type."

    trans = source.transform
    out = trans['shader'].out

    if out == 'image' or out == 'image_rgb':
        keys = source.data.keys()
        m = dict(zip(keys, keys))
        m['x_range'] = Range1d(start=0, end=0)
        m['y_range'] = Range1d(start=0, end=0)
        return m
    elif out == 'multi_line':
        keys = source.data.keys()
        m = dict(zip(keys, keys))
        m['line_color'] = 'colors'
        return m

    else:
        raise ValueError("Unrecognized shader output type %s" % out)


def make_glyphset(xcol, ycol, datacol, glyphspec, transform):
    # TODO: Do more detection to find if it is an area implantation.
    #       If so, make a selector with the right shape pattern and use a point shaper
    shaper = _shaper(glyphspec, transform['points'])
    if isinstance(shaper, glyphset.ToPoint):
        points = np.zeros((len(xcol), 4), order="F")
        points[:, 0] = xcol
        points[:, 1] = ycol
        glyphs = npg.Glyphset(points, datacol)
    else:
        glyphs = glyphset.Glyphset([xcol, ycol], datacol,
                                   shaper, colMajor=True)
    return glyphs


def _generate_render_state(plot_state):
    data_x_span = float(_span(plot_state['data_x']))
    data_y_span = float(_span(plot_state['data_y']))

    data_x_span = math.ceil(data_x_span * 100) / 100.0
    data_y_span = math.ceil(data_y_span * 100) / 100.0

    return {'x_span': data_x_span,
            'y_span': data_y_span}


def downsample(raw_data, data_source, glyph, plot_state, render_state, auto_bounds):
    _loadAR()  # Must be called before any attempts to use AR proper

    # XXX: transform['glyphspec'] is really a glyph
    data = raw_data
    transform = data_source.transform
    # the glyph which is passed in is the glyph asking for this computation
    # the glyph that is stored on the data source is the original glyph that
    # generated the data source (so if original is square, then new is image_rgba)
    glyphspec = get_glyphspec(transform['glyphspec'])
    xcol = glyphspec['x']['field']
    ycol = glyphspec['y']['field']
    datacol = _datacolumn(glyphspec)

    if render_state == _generate_render_state(plot_state):
        logger.info("Skipping update; render state unchanged")
        return {'render_state': "NO UPDATE"}

    # Translate the resample parameters to server-side rendering....
    # TODO: Do more to preserve the 'natural' data form and have make_glyphset build the 'right thing' (tm)
    if not isinstance(data, dict):
        columns = [xcol, ycol] + ([datacol] if datacol else [])
        data = data[columns]

    xcol = data[xcol]
    ycol = data[ycol]
    datacol = data[datacol] if datacol else np.zeros_like(xcol)
    glyphs = make_glyphset(xcol, ycol, datacol, glyphspec, transform)
    shader = transform['shader']

    if shader.out == "image" or shader.out == "image_rgb":
        rslt = downsample_image(xcol, ycol, glyphs, transform, plot_state, auto_bounds)
    elif shader.out == "multi_line":
        rslt = downsample_line(xcol, ycol, glyphs, transform, plot_state, auto_bounds)
    else:
        raise ValueError("Unhandled shader output '{0}.".format(shader.out))

    rslt['render_state'] = _generate_render_state(plot_state)

    return rslt


def downsample_line(xcol, ycol, glyphs, transform, plot_state, auto_bounds):
    # todo handle flipped axes (start < end)
    if auto_bounds:
        plot_state['data_x'].start = xcol.min()
        plot_state['data_x'].end = xcol.max()
        plot_state['data_y'].start = ycol.min()
        plot_state['data_y'].end = ycol.max()
    #screen_x_span = float(_span(plot_state['screen_x']))  # todo: use these?
    #screen_y_span = float(_span(plot_state['screen_y']))
    #data_x_span = float(_span(plot_state['data_x']))  # todo: use these?
    #data_y_span = float(_span(plot_state['data_y']))
    shader = transform['shader']

    bounds = glyphs.bounds()
    plot_size = [bounds[2], bounds[3]]
    balancedZoom = transform.get('balancedZoom', False)

    vt = util.zoom_fit(plot_size, bounds, balanced=balancedZoom)

    lines = ar.render(glyphs,
                      transform['info'].reify(),
                      transform['agg'].reify(glyphset=glyphs),
                      shader.reify(),
                      plot_size, vt)

    result = {'data' : shader.reformat(lines, xcol.min(), ycol.min())}
    result['x_range'] = {'start': plot_state['data_x'].start,
                         'end': plot_state['data_x'].end}
    result['y_range'] = {'start': plot_state['data_y'].start,
                         'end': plot_state['data_y'].end}
    logger.info("Finished line-producing downsample")
    return result


def downsample_image(xcol, ycol, glyphs, transform, plot_state, auto_bounds):
    logger.info("Starting image-producing downsample")
    # todo handle flipped axes (start < end)
    if auto_bounds:
        plot_state['data_x'].start = xcol.min()
        plot_state['data_x'].end = xcol.max()
        plot_state['data_y'].start = ycol.min()
        plot_state['data_y'].end = ycol.max()
    screen_x_span = float(_span(plot_state['screen_x']))
    screen_y_span = float(_span(plot_state['screen_y']))
    data_x_span = float(_span(plot_state['data_x']))
    data_y_span = float(_span(plot_state['data_y']))
    shader = transform['shader']
    #balanced_zoom = transform.get('balancedZoom', False)  # todo: use this?

    bounds = glyphs.bounds()
    scale_x = data_x_span/screen_x_span
    scale_y = data_y_span/screen_y_span
    plot_size = (bounds[2]/scale_x, bounds[3]/scale_y)

    vt = util.zoom_fit(plot_size, bounds, balanced=False)
    (tx, ty, sx, sy) = vt

    image = ar.render(glyphs,
                      transform['info'].reify(),
                      transform['agg'].reify(glyphset=glyphs),
                      shader.reify(),
                      plot_size, vt)

    image = shader.reformat(image)
    """AR renders the WHOLE dataset - then it's up to the client
    to only render the proper subset.  The client should also not
    re-request on operations hat do not require re-renders (i.e. pans)
    however I broke that in the blaze refactor - TODO :  is to fix that
    """
    result = {'data' : {'image': [image],
                        'x': [xcol.min()],
                        'y': [ycol.min()],
                        'dw': [xcol.max() - xcol.min()],
                        'dh': [ycol.max() - ycol.min()]}}
    result['x_range'] = {'start': plot_state['data_x'].start,
                         'end': plot_state['data_x'].end}
    result['y_range'] = {'start': plot_state['data_y'].start,
                         'end': plot_state['data_y'].end}
    logger.info("Finished image-producing downsample")
    return result


def _datacolumn(glyphspec):
    """Search the glyphspec to determine what the data column is.
    Returns the column name or False if none was found

    Precedence:
    Type > Fill Color > Fill Alpha > Line Color > Line Alpha

    TODO: Support a tuple-like datacolumn
    TODO: Support direct AR override parameter to directly specify data column
    TODO: Line width?
    """

    def maybe_get(key):
        return (key in glyphspec
                and isinstance(glyphspec[key], dict)
                and glyphspec[key].get('field', False))
    return (maybe_get('type')
            or maybe_get('fill_color')
            or maybe_get('fill_alpha')
            or maybe_get('line_color')
            or maybe_get('line_alpha'))


def _span(r):
    """Distance in a Range1D"""
    end = r.end if r.end is not None else 0
    start = r.start if r.start is not None else 0
    return abs(end-start)


def _shaper(glyphspec, points):
    """Construct the AR shaper to match the given shape code."""
    glyphtype = glyphspec['type']

    tox = glyphset.idx(0)
    toy = glyphset.idx(1)

    if points:
        sizer = glyphset.const(1)
        return glyphset.ToPoint(tox, toy, sizer, sizer)

    # XXX: glyphspec['dataspec'] won't work anymore, because previously all dataspecs
    # were serialized no matter if dirty or not. Now only changed dataspecs are being
    # serialized. Use glyphs directly to fix this issue.

    if glyphtype == 'Square':
        size = glyphspec.get('size', {}).get('value', 1)
        sizer = glyphset.const(size)
        return glyphset.ToRect(tox, toy, sizer, sizer)
    if glyphtype == 'Circle':
        # XXX: what about 'size' and different units?
        size = glyphspec.get('radius', {}).get('value', 1)
        sizer = glyphset.const(size)
        return glyphset.ToRect(tox, toy, sizer, sizer)
    else:
        raise ValueError("Unrecogized shape, received '{0}'".format(glyphtype.lower()))


# -------------------- Recipes -----------------------

def hdalpha(plot, cats=None, palette=None, log=True, spread=0, **kwargs):
    """
    Produce a plot using high-definition alpha composition (HDAlpha).
    HDAlpha essentially makes a heatmap for each of a number of categories,
    with different colors for each category.  Then those colors are composed
    in a way that ensures that oversaturate on does not occur.
    This is a convenience function that encodes a common configuration,
    and parameters to support the most common variations.

    plot -- Plot to convert into an HDAlpha plot
    cats -- What are the expected categories?
                 Default is None, indicating that categories
                 should be determined automatically.
    palette -- What colors should be used. Colors are matched to categories in
              order, so the first color is associated with the first category.
              Default is a rainbow palette with 8 steps.
    spread -- How far (if any) should values be spread. Default is 0.
    long -- Log-transform each category prior to composing? Default is True.
    kwargs -- Further arguments passed on to replot for greater control.

    Note: Category to color association follows the rules of the HDAlpha
         operator.  Refer to that documentation for corner case resolution.
    """

    if cats is None:
        info = AutoEncode()
    else:
        info = Encode(cats=cats)

    if palette is None:
        palette = ["#e41a1c", "#377eb8", "#4daf4a", "#984ea3", "#ff7f00", "#ffff33", "#a65628", "#f781bf"]

    kwargs['points'] = kwargs.get('points', True)

    return replot(plot,
                  info=info,
                  agg=CountCategories(),
                  shader=Spread(factor=spread) + HDAlpha(colors=palette, log=log),
                  **kwargs)


def heatmap(plot,
            client_color=False,
            low=(255, 200, 200), high=(255, 0, 0),
            spread=0, transform="cbrt", **kwargs):
    """
    Produce a heatmap from a set of shapes.
    A heatmap is a scale of how often a single thing occurs.
    This is a convenience function that encodes a common configuration,
    and parameters to support the most common variations.


    plot -- Plot to convert into a heatmap
    low -- Low color of the heatmap.  Default is a light red
    high -- High color of the heatmap. Default is full saturation red.
    spread -- How far (if any) should values be spread. Default is 0.
    transform -- Apply a transformation before building a color ramp?
                 Understood values are 'cbrt', 'log', 'none' and None.
                 The default is 'cbrt', for cuberoot, an approximation of
                 perceptual correction on monochromatic scales.
    kwargs -- Further arguments passed on to replot for greater control.
    """

    transform = transform.lower() if transform is not None else None

    if client_color:
        shader = Id()
        kwargs['reserve_val'] = kwargs.get('reserve_val', 0)
    else:
        shader = InterpolateColor(low=low, high=high)

    if transform == "cbrt":
        shader = Cuberoot() + shader
    elif transform == "log":
        shader = Log() + shader
    elif transform == "none" or transform is None:
        pass
    else:
        raise ValueError("Unrecognized transform '{0}'".format(transform))

    if spread > 0:
        shader = Spread(factor=spread, shape="circle") + shader

    kwargs['points'] = kwargs.get('points', True)

    return replot(plot,
                  agg=Count(),
                  info=Const(val=1),
                  shader=shader,
                  **kwargs)


def contours(plot, palette=None, transform='cbrt', spread=0, **kwargs):
    """
    Create ISO contours from a given plot

    plot -- Plot to convert into iso contours
    spread -- How far (if any) should values be spread. Default is 0.
    palette -- What should the line colors be?
               The number of lines is determined by the number of colors.
    transform -- Apply a transformation before building the iso contours?
                 Understood values are 'cbrt', 'log', 'none' and None.
                 The default is 'cbrt', for cuberoot, an approximation of
                 perceptual correction for monochromatic scales.
    kwargs -- Further arguments passed on to replot for greater control.
    """
    if palette is None:
        palette = ["#C6DBEF", "#9ECAE1", "#6BAED6", "#4292C6", "#2171B5", "#08519C", "#08306B"]

    shader = Contour(levels=len(palette), palette=palette)

    if transform == "cbrt":
        shader = Cuberoot() + shader
    elif transform == "log":
        shader = Log() + shader
    elif transform == "none" or transform is None:
        pass
    else:
        raise ValueError("Unrecognized transform '{0}'".format(transform))

    if spread > 0:
        shader = Spread(factor=spread, shape="circle") + shader

    return replot(plot,
                  agg=Count(),
                  info=Const(val=1),
                  shader=shader,
                  **kwargs)
